import pickle
from typing import List, Optional, Tuple
import os
from modAL.models.learners import ActiveLearner
from pydantic.main import BaseModel
from icolos.core.workflow_steps.active_learning.base import ActiveLearningBase
from icolos.core.workflow_steps.step import _LE
from icolos.utils.enums.step_enums import (
    StepActiveLearningEnum,
)
from rdkit.Chem import PandasTools
from rdkit import Chem
import pandas as pd
from pandas.core.frame import DataFrame
import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error


_SALE = StepActiveLearningEnum()


class StepActiveLearning(ActiveLearningBase, BaseModel):
    """
    Class to run an active learning framework
    Primarily designed for building QSAR models using a physics based method (embedding + docking) as an oracle

    Takes the step conf for the oracle as an additional argument.  The step with these settings is run with the queried compounds at each stage of the active learning loop
    """

    def __init__(self, **data):
        super().__init__(**data)

    def _parse_library(
        self, lib_path: str, criteria: str = None
    ) -> Tuple[DataFrame, Optional[pd.Series]]:
        """Parse a virtual library to a dataframe

        :param str criteria: optional criteria to extract to df column from sdf tags, defaults to None
        :raises ValueError: Accepts only sdf or smi files
        :return Tuple[DataFrame, np.ndarray]: tuple containing datafarme of parsed data and scores, empty if no criteria provided
        """

        if lib_path.endswith("sdf"):
            # hold the lib in a pandas df
            library = PandasTools.LoadSDF(
                lib_path,
                smilesName=_SALE.SMILES,
                molColName=_SALE.MOLECULE,
                includeFingerprints=True,
                removeHs=False,
                embedProps=True,
            )
            library = self.construct_fingerprints(library)
        elif lib_path.endswith("smi"):
            mols, smiles = [], []
            suppl = Chem.rdmolfiles.SmilesMolSupplier(lib_path)
            for mol in suppl:
                mols.append(mol)
                smiles.append(Chem.MolToSmiles(mol))
            library = pd.DataFrame({_SALE.SMILES: smiles, _SALE.MOLECULE: mols})
            library = self.construct_fingerprints(library)
        elif lib_path.endswith("pkl"):
            library = pd.read_pickle(lib_path)
        else:
            raise ValueError(f"File {lib_path} must of of type smi, pkl or sdf")
        # randomly shuffle the compound rows
        library = library.sample(frac=1).reset_index(drop=True)
        scores = (
            pd.to_numeric(library[criteria].fillna(0)) if criteria is not None else None
        )

        return library, scores

    def _run_learning_loop(
        self,
        learner: ActiveLearner,
        lib: pd.DataFrame,
        tmp_dir: str,
        all_scores: List = None,
        top_1_idx: list = None,
        replica=0,
        fragment_lib: pd.DataFrame = None,
    ) -> None:
        """Main active learning loop, runs the train and query loop on the virtual lib

        :param ActiveLearner learner: Initialized modAL active learning object
        :param pd.DataFrame lib: dataframe containing compounds to screen
        :param str tmp_dir: tmpdir to run the calculations in
        :param list top_1_idx: optional list of of indices of top 1% compounds, for retrospective performance eval, defaults to None
        :param int replica: id of the experiment replica, defaults to 0
        :param pd.DataFrame fragment_lib: for non-natural amino acids, pass a df containing the non-natural fragments, only for protein FEP and residue scanning oracles. defaults to None
        """

        def query_surrogate(prev_idx: List, warmup: bool) -> List:
            query_idx, _ = learner.query(
                X,
                n_instances=batch_size,
                previous_idx=prev_idx,
                warmup=warmup,
                epsilon=epsilon,
            )
            queried_compounds_per_epoch.append([int(i) for i in query_idx])
            return list(query_idx)

        def get_precomputed_scores():
            self._logger.log("Retrieving scores from precomputed data...", _LE.INFO)
            scores = np.array(
                [
                    float(lib.iloc[int(idx)][self.settings.additional[_SALE.CRITERIA]])
                    for idx in query_idx
                ],
                dtype=np.float32,
            )
            return scores

        def get_scores_from_oracle():
            self._logger.log(
                f"Querying oracle with {len(query_idx)} compounds", _LE.INFO
            )

            scores = self.query_oracle(
                query_compounds,
                fragment_lib=fragment_lib,
                oracle_type=oracle_type,
                round=rnd,
            )

            return scores

        def compute_rmsd() -> float:
            """Generate rmsd values"""
            predictions = learner.predict(X)
            return np.sqrt(mean_squared_error(all_scores, predictions))

        n_instances = (
            int(self._get_additional_setting(_SALE.FRACTION_PER_EPOCH) * len(lib))
            if self._get_additional_setting(_SALE.FRACTION_PER_EPOCH) is not None
            else int(self._get_additional_setting(_SALE.BATCH_SIZE))
        )
        rounds = (
            int(
                (self._get_additional_setting(_SALE.MAX_SAMPLED_FRACTION) * len(lib))
                / n_instances
            )
            if self._get_additional_setting(_SALE.MAX_SAMPLED_FRACTION) is not None
            else int(self._get_additional_setting(_SALE.N_ROUNDS))
        )
        queried_compound_idx = []
        queried_compounds_per_epoch = []
        fraction_top1_hits_per_epoch = []
        key = _SALE.MORGAN_FP
        warmup_period = self._get_additional_setting(_SALE.WARMUP, default=1)
        X = np.array(list(lib[key]))
        oracle_type = self._get_additional_setting(_SALE.ORACLE_TYPE, default="docking")
        epsilon = float(self._get_additional_setting(_SALE.EPSILON, default=0.0))
        self._logger.log(
            f"Running {rounds} rounds of {n_instances} componds, replica {replica}",
            _LE.DEBUG,
        )
        for rnd in range(rounds):
            # are we still in the warmup period?
            warmup = rnd < warmup_period
            batch_size = (
                int(
                    n_instances
                    * self._get_additional_setting(_SALE.INIT_SAMPLE_FACTOR, default=1)
                )
                if rnd == 0
                else n_instances
            )
            query_idx = query_surrogate(queried_compound_idx, warmup)
            if all_scores is not None:
                rmsd = compute_rmsd() if rnd > 0 else np.inf

            self._logger.log(f"queried compound indices are {query_idx}", _LE.DEBUG)
            queried_compound_idx += query_idx
            query_compounds = [lib.iloc[int(idx)] for idx in query_idx]

            # check evaluation mode or not
            if self._get_additional_setting(_SALE.EVALUATE, default=False):
                scores = get_precomputed_scores()
            else:
                scores = get_scores_from_oracle()
            if self.settings.additional["model"] == "ffnn":
                scores = scores.reshape(-1, 1)

            self._logger.log("Fitting surrogate with new data...", _LE.INFO)
            new_data = np.array([compound[key] for compound in query_compounds])
            learner.teach(new_data, scores, only_new=False)
            # calculate percentage of top-1% compounds queried
            if top_1_idx:
                hit_count = 0
                for idx in top_1_idx:
                    if idx in queried_compound_idx:
                        hit_count += 1
                hits_queried = hit_count / len(top_1_idx) * 100
                print(f"Round {rnd}: top1%: {hits_queried}%, rmsd: {rmsd}")
                fraction_top1_hits_per_epoch.append(hits_queried)
                if hits_queried > 99:
                    break

            df = lib.iloc[query_idx]
            # save the scores in the dataframe
            df["oracle_score"] = scores
            df.to_pickle(
                os.path.join(tmp_dir, f"enriched_lib_rep_{replica}_round_{rnd+1}.pkl")
            )
            PandasTools.WriteSDF(
                df,
                os.path.join(tmp_dir, f"compounds_{replica}_{rnd}.sdf"),
                molColName=_SALE.MOLECULE,
            )
            # pickle the model
            with open(os.path.join(tmp_dir, f"model_{replica}_{rnd}.pkl"), "wb") as f:
                pickle.dump(learner.estimator, f)
            if (
                self._get_additional_setting(_SALE.DYNAMIC_STOP, default=False) is True
                and rnd > 4
            ):
                # average of last three top-1 fractions
                rolling_avg = np.mean(fraction_top1_hits_per_epoch[-5:-1])
                if fraction_top1_hits_per_epoch[-1] - rolling_avg < 0.01:
                    self._logger.log(
                        f"Fraction top1 hits converged to 0.01, stopping...", _LE.INFO
                    )
                    break

        # create a results dataframe to store per-epoch properties
        self._logger.log("Generating results dataframe...", _LE.DEBUG)

        # we have a list of compound IDs per epoch, we want the transpose
        try:
            queried_compounds_per_position = np.array(queried_compounds_per_epoch).T
            col_dict = {
                "epoch": [i for i in range(len(fraction_top1_hits_per_epoch))],
                "top_1%_per_epoch": fraction_top1_hits_per_epoch,
            }
            for idx, row in enumerate(queried_compounds_per_position):
                col_dict[f"seq_idx_pos_{idx}"] = list(row)
            resuts_df = pd.DataFrame(col_dict)
            resuts_df.to_csv(os.path.join(tmp_dir, f"results_rep_{replica}.csv"))
        except ValueError as e:
            self._logger.log(
                f"Failed to generate results csv, error was {e}", _LE.WARNING
            )
        # return the enriched df
        enriched_lib = lib.iloc[queried_compound_idx]
        enriched_lib.to_pickle(os.path.join(tmp_dir, f"enriched_lib_rep_{replica}.pkl"))

    def execute(self):
        tmp_dir = self._make_tmpdir()
        print(tmp_dir)
        criteria = (
            self._get_additional_setting(_SALE.CRITERIA)
            if self._get_additional_setting(_SALE.EVALUATE, default=False)
            else None
        )
        lib, scores = self._parse_library(
            self._get_additional_setting(_SALE.VIRTUAL_LIB), criteria=criteria
        )
        lib.to_pickle(os.path.join(tmp_dir, "starting_lib.pkl"))
        if isinstance(scores, pd.Series):
            top_1_percent = int(0.01 * len(scores))
            # this assumes lowest is best
            top_1_idx = np.argpartition(scores, top_1_percent)[:top_1_percent]
            self._logger.log(
                f"Using {len(top_1_idx)} top compounds for validation", _LE.INFO
            )
            print(f"top scores are, {scores[top_1_idx]}")
            for idx, score in enumerate(scores[top_1_idx]):
                print(idx, score)
        else:
            top_1_idx = []

        # load fragment lib if provided
        if self._get_additional_setting(_SALE.FRAGMENTS) is not None:
            print("Loading fragment lib")
            fragments_library, _ = self._parse_library(
                self._get_additional_setting(_SALE.FRAGMENTS)
            )
            fragments_library.to_pickle(os.path.join(tmp_dir, "fragments.pkl"))
        else:
            fragments_library = None
        replicas = self._get_additional_setting(_SALE.REPLICAS, default=1)
        for replica in range(replicas):
            learner = self._initialize_learner()

            self._logger.log("Evaluating virtual library...", _LE.DEBUG)
            self._run_learning_loop(
                learner=learner,
                lib=lib,
                fragment_lib=fragments_library,
                all_scores=scores,
                tmp_dir=tmp_dir,
                top_1_idx=list(top_1_idx),
                replica=replica,
            )



from icolos.utils.execute_external.execute import ExecutorBase


class CressetExecutor(ExecutorBase):
    """For the execution of Cresset binaries binary."""

    def __init__(self, prefix_execution=None, binary_location=None):
        super().__init__(
            prefix_execution=prefix_execution, binary_location=binary_location
        )

    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):
        # check, whether a proper executable is provided
        # if command not in [EE.OMEGA]:
        #     raise ValueError(
        #         "Parameter command must be an dictionary of the internal Omega executable list."
        #     )

        return super().execute(
            command=command,
            arguments=arguments,
            check=check,
            location=None,
            pipe_input=pipe_input,
        )

    def is_available(self):
        # try:
        #     result = self.execute(
        #         command=EE.OMEGA, arguments=[EE.OMEGA_HELP], check=True
        #     )

        #     if EE.OMEGA_HELP_IDENTIFICATION_STRING in result.stderr:
        #         return True
        #     return False
        # except Exception as e:
        #     return False
        pass



class LoggingConfigEnum:

    # set levels (for now, they match to the "logging" default ones)
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    EXCEPTION = "exception"

    # paths to the configuration JSONs that are shipped with Icolos
    PATH_CONFIG_DEFAULT = "src/icolos/config/logging/default.json"
    PATH_CONFIG_VERBOSE = "src/icolos/config/logging/verbose.json"
    PATH_CONFIG_DEBUG = "src/icolos/config/logging/debug.json"
    PATH_CONFIG_TUTORIAL = "src/icolos/config/logging/tutorial.json"

    # high-level loggers defined in the configurations
    LOGGER_IO = "io"
    LOGGER_STEP = "step"
    LOGGER_AGENT = "agent"
    LOGGER_ENTRYPOINT = "entrypoint"
    LOGGER_BLANK = "blank"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")



from icolos.core.containers.generic import GenericData
import os
import tempfile
import re
import numpy as np
from copy import deepcopy
from typing import List

from icolos.core.workflow_steps.calculation.base import StepCalculationBase
from icolos.utils.enums.program_parameters import PantherEnum
from icolos.utils.enums.step_enums import StepPantherEnum
from icolos.utils.execute_external.execute import Executor
from icolos.core.workflow_steps.step import _LE
from pydantic import BaseModel
from icolos.utils.general.files_paths import attach_root_path

_SPE = (
    StepPantherEnum()
)  # hold the constants to access the relevant value from initialised **data
_PE = PantherEnum()  # hold the program settings


class StepPanther(StepCalculationBase, BaseModel):

    negative_images: List = []

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=Executor)

    def _write_panther_config_file(self, tmp_dir):
        if not self.settings.additional[_SPE.PANTHER_CONFIG_FILE]:
            self._logger.log("No config file specified, using default.", _LE.INFO)
            panther_config = attach_root_path(
                "/icolos/config/panther/default_panther.in"
            )

        elif not os.path.isfile(self.settings.additional[_SPE.PANTHER_CONFIG_FILE]):
            self._logger.log(
                f"File not found for the provided panther config file path: {self.settings.additional[_SPE.PANTHER_CONFIG_FILE]}",
                _LE.ERROR,
            )
            raise FileNotFoundError(
                f"The specified panther config file was not found {self.settings.additional[_SPE.PANTHER_CONFIG_FILE]}"
            )

        else:
            panther_config = self.settings.additional[_SPE.PANTHER_CONFIG_FILE]

        with open(panther_config, "r") as f:
            panther_config = f.read()

        # add the parameter absolute paths to the angle, etc. file specifications
        update_dictionary = deepcopy(self.settings.additional[_SPE.FIELDS])
        update_dictionary = self._add_ligand_centroid_coordinates(update_dictionary)
        update_dictionary = self._add_parameter_locations_to_replacement_fields(
            update_dictionary
        )
        # update the configuration and write it to a file
        panther_config = self._modify_panther_config_file(
            panther_config, update_dictionary
        )

        with open(os.path.join(tmp_dir, "panther_config.in"), "w") as f:
            f.write(panther_config)

    def _add_parameter_locations_to_replacement_fields(
        self, update_dictionary: dict
    ) -> dict:
        # in case not specified (which is the main use case), use the default libraries for charges etc. that should
        # reside in the same folder as the python entry-point "panther.py"; setting absolute paths here, allows to
        # execute PANTHER for any input in any given folder
        for key, value in _SPE.FIELDS_PARAMETERS_LIB.items():
            if key not in update_dictionary.keys():
                update_dictionary[key] = os.path.join(
                    self.settings.additional[_SPE.PANTHER_LOCATION], value
                )
        return update_dictionary

    def _add_ligand_centroid_coordinates(self, update_dict: dict) -> dict:
        coordinates = self._calculate_ligand_centroid(
            self.settings.additional[_SPE.FIELDS][_SPE.FIELD_KEY_PDB_FILE]
        )
        update_dict[_SPE.FIELD_KEY_COORDINATES] = coordinates
        return update_dict

    def _calculate_ligand_centroid(self, file: str) -> str:
        with open(file, "r") as f:
            file_lines = f.readlines()
        file_lines = [
            line for line in file_lines if "X   0" in line and len(line.split()) > 5
        ]
        if not file_lines:
            self._logger.log(
                "No lines corresponding to the ligand found! Centroid will not be correct",
                _LE.WARNING,
            )
        a = np.genfromtxt(file_lines, usecols=[6, 7, 8], skip_header=1)
        avg = list(a.mean(axis=0))
        avg = [str(i) for i in avg]
        return " ".join(avg)

    def _modify_panther_config_file(
        self, config_file: str, update_dictionary: dict
    ) -> str:
        for key, value in update_dictionary.items():
            pattern = rf"({key}.*:: ).*"
            pattern = re.compile(pattern)
            config_file = re.sub(pattern, rf"\1 {value}", config_file)
        return config_file

    def _execute_backend(self, tmp_dir):
        arguments = [
            os.path.join(
                self.settings.additional[_SPE.PANTHER_LOCATION], _PE.PANTHER_ENTRYPOINT
            ),
            os.path.join(tmp_dir, _PE.PANTHER_CONFIG),
            os.path.join(tmp_dir, _PE.PANTHER_OUTPUT_FILE),
        ]
        self._backend_executor.execute(
            command=_PE.PANTHER_PTYHON2, arguments=arguments, check=True
        )

    def _parse_panther_output(self, tmp_dir):
        try:
            with open(os.path.join(tmp_dir, _PE.PANTHER_OUTPUT_FILE), "r") as f:
                data = f.read()
                self.data.generic.add_file(
                    GenericData(file_name=_PE.PANTHER_OUTPUT_FILE, file_data=data)
                )
        except FileNotFoundError:
            self._logger.log(
                f"No panther output file was produced for step {self.step_id}, subsequent steps that depend on the negative image will fail.",
                _LE.WARNING,
            )

    def execute(self):
        tmp_dir = self._make_tmpdir()
        self._write_panther_config_file(tmp_dir)
        self._execute_backend(tmp_dir)
        self._logger.log("Executed PANTHER and obtained negative image.", _LE.INFO)
        self._logger.log(
            f"Calculated negative image for configuration file in {tmp_dir}.", _LE.DEBUG
        )
        self._parse_panther_output(tmp_dir)
        self._remove_temporary(tmp_dir)



from copy import deepcopy
from icolos.core.containers.generic import GenericContainer, GenericData
import json
import pandas as pd
from rdkit import Chem
from icolos.core.containers.gmx_state import GromacsState

from icolos.loggers.base_logger import BaseLogger
from icolos.utils.enums.input_enums import InputEnum
from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.enums.write_out_enums import WriteOutEnum
from icolos.utils.general.icolos_exceptions import StepFailed
from icolos.utils.smiles import to_smiles
from icolos.utils.general.files_paths import infer_input_type

from typing import List, Any
from pydantic import BaseModel

from icolos.core.step_utils.input_merger import InputMerger, StepMerge
from icolos.core.containers.compound import Enumeration, Compound, Conformer
from icolos.utils.enums.step_enums import StepBaseEnum, StepGromacsEnum
import os
from tempfile import mkdtemp
import requests

_SBE = StepBaseEnum
_LE = LoggingConfigEnum()
_WE = WriteOutEnum()
_SGE = StepGromacsEnum()
_IE = InputEnum()


class StringPath(str):
    def __new__(cls, content):
        return super().__new__(cls, content)


class StringFile(str):
    def __new__(cls, content):
        return super().__new__(cls, content)


class StepData(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    compounds: List[Compound] = []
    generic: GenericContainer = GenericContainer()
    gmx_state: GromacsState = GromacsState()


class StepCSVInputColumnParameters(BaseModel):
    smiles: str
    names: str = None


class StepInputEnforceIDs(BaseModel):
    compound_ids: List = None
    enumeration_ids: List = None


class StepInputSource(BaseModel):
    source: str
    source_type: str = None
    source_field: str = _IE.SOURCE_FIELD_COMPOUNDS
    target_field: str = _IE.SOURCE_FIELD_COMPOUNDS
    extension: str = None
    format: str = None
    delimiter: str = ","
    columns: StepCSVInputColumnParameters = None
    enforce_ids: StepInputEnforceIDs = None


class StepInputParameters(BaseModel):

    compounds: List[StepInputSource] = []
    generic: List[StepInputSource] = []
    gmx_state: StepInputSource = {}
    perturbation_map: List[StepInputSource] = None
    merge: StepMerge = StepMerge()
    work_dir: str = None


class InputPreparator(BaseModel):
    workflow: Any = None
    logger: BaseLogger = None

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True

    def __init__(self, **data):
        super().__init__(**data)

    def generate_input(self, step_input: StepInputParameters, step_type):
        compounds = self._generate_compound_input(step_input)
        generic = (
            self._generate_generic_input(step_input, step_type)
            if step_input.generic
            else GenericContainer()
        )
        gmx_state = (
            self._generate_gmx_state_input(step_input)
            if step_input.gmx_state
            else GromacsState()
        )
        # Instruct the step to run in a specific workdir, e.g. from a previously failed job or to execute a few related steps in the same dir
        if step_input.work_dir is not None:
            if os.path.isdir(step_input.work_dir):
                work_dir = step_input.work_dir
                self.logger.log(
                    f"Found specified work dir at {step_input.work_dir}", _LE.DEBUG
                )
                # now check whether this needs attaching to the workflow for the rest of the steps
                if self.workflow.header.global_settings.single_directory:
                    self.workflow.workflow_data.work_dir = work_dir
                    self.logger.log(
                        f"Setting workdir  at {step_input.work_dir} to the workflow's workdir",
                        _LE.DEBUG,
                    )
            else:
                # last resort, if a previous step_id has been passed, get the work_dir from there

                work_dir = self.workflow.find_step_by_step_id(
                    step_input.work_dir
                ).work_dir
        elif (
            self.workflow is not None
            and self.workflow.header.global_settings.single_directory
        ):
            # Entire workflow running in a single dir (pmx), either generate one for the first
            # step or use the already generated dir
            work_dir = self._get_workflow_workdir()
        else:
            work_dir = None
        return (
            StepData(compounds=compounds, generic=generic, gmx_state=gmx_state),
            work_dir,
        )

    def _get_workflow_workdir(self):
        # check whether the workflow already has one attached, otherwise create one
        if self.workflow.workflow_data.work_dir is not None and os.path.isdir(
            self.workflow.workflow_data.work_dir
        ):
            return self.workflow.workflow_data.work_dir
        else:
            tmp_dir = mkdtemp()
            self.workflow.workflow_data.work_dir = tmp_dir
            self.logger.log(f"Set workflow's tmpdir to {tmp_dir}", _LE.DEBUG)
            return tmp_dir

    def _generate_compound_input(self, step_input: StepInputParameters) -> List:
        compounds = []
        for inp in step_input.compounds:
            if inp.target_field == _IE.TARGET_FIELD_COMPOUNDS:
                buffer = []
                if inp.source_type == _SBE.INPUT_SOURCE_TYPE_FILE:
                    buffer.append(self._read_compound_input_from_file(inp))
                elif inp.source_type == _SBE.INPUT_SOURCE_TYPE_STEP:
                    buffer.append(self._read_compound_input_from_step(inp))
                elif inp.source_type == _SBE.INPUT_SOURCE_TYPE_STRING:
                    buffer.append(self._read_input_from_string(inp))
                else:
                    raise ValueError(
                        f"Source type {inp.source_type} for compound input unsupported - abort."
                    )
                if inp.target_field == _IE.SOURCE_FIELD_COMPOUNDS:
                    # note: no unrolling here!
                    compounds = compounds + buffer

            elif inp.target_field == _IE.TARGET_FIELD_CONFORMERS:
                if inp.source_type == _SBE.INPUT_SOURCE_TYPE_FILE:
                    compounds = compounds + self._read_conformers_input_from_file(inp)
        if len(compounds) > 0:
            compounds = self._apply_compound_merger(step_input, compounds)
        return compounds

    def _generate_gmx_state_input(
        self, step_input: StepInputParameters
    ) -> GromacsState:
        input_step = self.workflow.find_step_by_step_id(step_input.gmx_state.source)
        return deepcopy(input_step.data.gmx_state)

    def _generate_generic_input(
        self, step_input: StepInputParameters, step_type
    ) -> GenericContainer:
        generic = GenericContainer()
        for inp in step_input.generic:
            files = self._read_data_to_generic(inp)
            generic.add_files(files)
        return generic

    def _read_data_to_generic(self, inp: StepInputSource):
        ext = inp.extension
        if inp.source_type == _SBE.INPUT_SOURCE_TYPE_FILE or os.path.isfile(inp.source):
            assert os.path.isfile(inp.source)
            try:
                with open(inp.source, "r") as f:
                    data = f.read()
            except UnicodeDecodeError:
                with open(inp.source, "rb") as f:
                    data = f.read()
            file = GenericData(inp.source.split("/")[-1], data)
            return [file]
        elif inp.source_type == _SBE.INPUT_SOURCE_TYPE_URL or inp.source.startswith(
            "http"
        ):
            data = self._get_pdb_file_from_api(inp.source)
            file_name = inp.source.split("/")[-1].split(".")[0] + "." + inp.extension
            file = GenericData(file_name=file_name, file_data=data)
            return [file]
        elif inp.source_type == _SBE.INPUT_SOURCE_TYPE_DIR or os.path.isdir(inp.source):
            assert os.path.isdir(inp.source)
            file = GenericData(
                file_data=inp.source,
                file_name=inp.source.split("/")[-1],
                extension=inp.extension,
            )
            return [file]
        else:
            # fall back on step source type
            input_step = self.workflow.find_step_by_step_id(inp.source)
            files = input_step.data.generic.get_files_by_extension(ext)

            # special case for itp and ndx files, these are included in the topol file so are never arguments
            if ext in ["itp", "ndx"]:
                return files

            if len(files) == 1:
                file = files[0]
                return [file]
            # else use the argument method
            else:
                # this introduces a manual check on which file the user wants if there are multiple
                file = input_step.data.generic.get_argument_by_extension(
                    ext, rtn_file_object=True
                )
                return [file]

    def _get_pdb_file_from_api(self, pdb_url: str):
        response = self._get_request(pdb_url)
        if response is None or not response.ok:
            return None
        return response.text

    def _get_request(self, url, max_tries=5):
        trials = 0
        while trials < max_tries:
            response = requests.get(url)
            if response.status_code == 200:
                return response

    def _apply_compound_merger(
        self, step_input: StepInputParameters, compounds: List[Compound]
    ) -> List[Compound]:
        merger = InputMerger(step_input.merge)
        compounds = merger.unroll_compounds(compounds)
        if not any(
            [
                True
                for compound in step_input.compounds
                if compound.enforce_ids is not None
            ]
        ):
            compounds = merger.merge(compounds=compounds)

        if len(compounds) == 0 and self.logger is not None:
            self.logger.log(
                "Input list of compounds is empty, this is likely an error.",
                _LE.WARNING,
            )
        return compounds

    def _read_compound_input_from_step(self, inp: StepInputSource):
        input_step = self.workflow.find_step_by_step_id(inp.source)
        return input_step.clone_compounds()

    def _read_conformers_input_from_file(self, inp: StepInputSource):
        # set up path to input file and extract the input format
        input_format = inp.format
        if input_format is None and self.logger is not None:
            self.logger.log(
                "No input format specified, will try to infer type (not recommended).",
                _LE.WARNING,
            )
            input_format = infer_input_type(inp.source)
        input_format = input_format.upper()

        # call the respective loading function
        if input_format == _SBE.FORMAT_SDF:
            compound = Compound(compound_number=0)
            enumeration = Enumeration()
            for mol_id, mol in enumerate(
                Chem.SDMolSupplier(inp.source, removeHs=False)
            ):
                conformer = Conformer(conformer=mol, enumeration_object=enumeration)
                enumeration.add_conformer(conformer=conformer, auto_update=True)
            compound.add_enumeration(enumeration, auto_update=True)
            return [compound]
        else:
            raise ValueError(
                f"At the moment, input format {input_format} is not supported."
            )

    def _read_compound_input_from_file(self, inp: StepInputSource):
        # set up path to input file and extract the input format
        input_format = inp.format
        if input_format is None and self.logger is not None:
            self.logger.log(
                "No input format specified, will try to infer type (not recommended).",
                _LE.WARNING,
            )
            input_format = infer_input_type(inp.source)
        input_format = input_format.upper()

        # call the respective loading function
        if input_format == _SBE.FORMAT_SDF:
            result = self._read_in_SDF_file(inp)
        elif input_format == _SBE.FORMAT_CSV:
            result = self._read_in_CSV_file(inp)
        elif input_format == _SBE.FORMAT_SMI:
            result = self._read_in_SMI_file(inp)
        elif input_format == _SBE.FORMAT_JSON:
            result = self._read_in_JSON_file(inp)
        else:
            raise ValueError(
                f"At the moment, input format {input_format} is not supported."
            )

        # apply ID enforcement, if specified
        return self._enforce_ids(result, inp)

    def _read_input_from_string(self, inp: StepInputSource) -> List[Compound]:
        # the strings must be separated by a semi-colon (';'); they may have names in front separated by a colon (':')
        elements = inp.source.split(";")
        list_compounds = []
        for line_id, line in enumerate(elements):
            # it could be, that names are part of the elements; otherwise use the number
            # remove trailing or preceding white spaces
            parts = [x.strip() for x in line.split(":")]
            if len(parts) == 2:
                compound = Compound(name=parts[0], compound_number=line_id)
                smile = parts[1]
            else:
                compound = Compound(name=str(line_id), compound_number=line_id)
                smile = parts[0]
            enumeration = Enumeration(smile=smile, original_smile=smile)
            compound.add_enumeration(enumeration, auto_update=True)
            list_compounds.append(compound)

        # apply ID enforcement, if specified
        return self._enforce_ids(list_compounds, inp)

    def _read_in_SDF_file(self, inp: StepInputSource) -> List[Compound]:
        """Loads compounds from an SDF file.  Icolos naming conventions will be parsed, if provided, such that compound structures can be reconstructed directly from a previously written sdf

        :param StepInputSource inp: input block from the config file
        :return List[Compound]: List of initialized compounds.  If Icolos-type naming was used, enumerations and conformers will be generated and attached automatically
        """

        def _get_existing_enumeration(comp_id, enum_id):
            comp = _get_existing_compound(comp_id)
            for enum in comp.get_enumerations():
                if enum.get_enumeration_id() == int(enum_id):
                    return enum
            raise ValueError

        def _get_existing_compound(idx):
            for comp in list_compounds:
                if int(idx) == comp.get_compound_number():
                    return comp
            raise ValueError

        list_compounds = []
        compound_number = 0
        icolos_naming = True
        # Parses compounds following the Icolos naming convention of Compound:Enumeration:Conformer to reconstruct the compound object
        for mol in Chem.SDMolSupplier(inp.source, removeHs=False):
            new_compound = False
            new_enumeration = False
            mol_name = mol.GetProp(_WE.RDKIT_NAME)
            # assuming the mol name follows Icolos conventions
            try:
                id_parts = mol_name.split(":")
                comp_id = id_parts[0]
                enum_id = id_parts[1]

            except IndexError:
                icolos_naming = False
                comp_id = mol_name
                enum_id = 0

            if icolos_naming:
                # reconstruct compound objects
                try:
                    # try to find an existing compound with the correct name
                    compound = _get_existing_compound(idx=comp_id)
                except ValueError:
                    # the compound does not yet exist, create the object
                    new_compound = True
                    try:
                        # if we have standard icolos compound naming
                        comp_num = int(comp_id)
                    except ValueError:
                        # some other naming scheme
                        comp_num = compound_number
                    compound = Compound(name=comp_id, compound_number=comp_num)
                try:
                    # check whether the enumeration exists
                    enumeration = _get_existing_enumeration(comp_id, enum_id)
                except ValueError:
                    new_enumeration = True
                    enumeration = Enumeration(
                        smile=to_smiles(mol),
                        molecule=mol,
                        original_smile=to_smiles(mol),
                    )

                if len(id_parts) == 3:
                    # i.e. 0:0:0, we have a conformer
                    conf = Conformer(
                        conformer=mol,
                        enumeration_object=enumeration,
                        conformer_id=int(id_parts[2]),
                    )
                    enumeration.add_conformer(conf, auto_update=True)
                if new_enumeration:
                    compound.add_enumeration(enumeration, auto_update=True)
                if new_compound:
                    list_compounds.append(compound)

            else:
                # if non-standard naming conventions, simply load each mol into a new compound object, with single enum/conf
                compound = Compound(name=comp_id, compound_number=compound_number)
                enum = Enumeration(
                    smile=to_smiles(mol),
                    molecule=mol,
                    original_smile=to_smiles(mol),
                    enumeration_id=0,
                )
                enum.add_conformer(
                    Conformer(conformer=mol, enumeration_object=mol, conformer_id=0),
                    auto_update=True,
                )
                compound.add_enumeration(enumeration=enum)
                list_compounds.append(compound)

            compound_number += 1
        return list_compounds

    def _read_in_SMI_file(self, inp: StepInputSource) -> List[Compound]:
        list_compounds = []
        with open(inp.source, "r") as f:
            # while the SMI file definition requires a name (separated by blanks) for each line
            # as well, assume that this might not be present
            lines = [line.rstrip() for line in f.readlines()]
        for line_id, line in enumerate(lines):
            if line == "":
                continue

            parts = line.split()
            if len(parts) == 2:
                compound = Compound(name=parts[1], compound_number=line_id)
            else:
                compound = Compound(name=str(line_id), compound_number=line_id)
            enumeration = Enumeration(smile=parts[0], original_smile=parts[0])
            compound.add_enumeration(enumeration, auto_update=True)
            list_compounds.append(compound)
        return list_compounds

    def _read_in_JSON_file(self, inp: StepInputSource) -> List[Compound]:
        list_compounds = []

        # load input
        with open(inp.source, "r") as f:
            inp_json = f.read().replace("\r", "").replace("\n", "")
            inp_dict = json.loads(inp_json)

        comp_id = 0
        for name, smile in zip(inp_dict[_IE.JSON_NAMES], inp_dict[_IE.JSON_SMILES]):
            compound = Compound(name=name, compound_number=comp_id)
            enumeration = Enumeration(smile=smile, original_smile=smile)
            compound.add_enumeration(enumeration, auto_update=True)
            list_compounds.append(compound)
            comp_id += 1

        return list_compounds

    def _read_in_CSV_file(self, inp: StepInputSource) -> List[Compound]:
        list_compounds = []
        delimiter = inp.delimiter
        data = pd.read_csv(inp.source, delimiter=delimiter)

        smiles_column = inp.columns.smiles
        if smiles_column not in list(data.columns):
            raise StepFailed(
                f"Column name for the smiles either not set or not found in input CSV."
            )

        # deal with names (if specified)
        names_column = inp.columns.names
        if names_column is None:
            names_compounds = None
        else:
            if names_column not in list(data.columns):
                raise StepFailed(
                    f"Specified column name ({names_column}) for the names either not found in input CSV."
                )
            else:
                names_compounds = [
                    str(name).strip() for name in data[names_column].tolist()
                ]

        # build the compounds
        smiles = [str(line).strip() for line in data[smiles_column].tolist()]
        for number in range(len(smiles)):
            if names_compounds is not None:
                compound = Compound(
                    name=names_compounds[number], compound_number=number
                )
            else:
                compound = Compound(name=str(number), compound_number=number)
            enumeration = Enumeration(
                smile=smiles[number], original_smile=smiles[number]
            )
            compound.add_enumeration(enumeration, auto_update=True)
            list_compounds.append(compound)
        return list_compounds

    def _enforce_ids(
        self, compounds: List[Compound], inp: StepInputSource
    ) -> List[Compound]:
        if inp.enforce_ids is not None:
            if inp.enforce_ids.compound_ids is not None:
                for comp_idx, comp in enumerate(compounds):
                    comp.set_compound_number(
                        int(inp.enforce_ids.compound_ids[comp_idx])
                    )

            # set enumeration ids
            enum_id_idx = 0
            if inp.enforce_ids.enumeration_ids is not None:
                for comp in compounds:
                    for enum in comp.get_enumerations():
                        enum.set_enumeration_id(
                            int(inp.enforce_ids.enumeration_ids[enum_id_idx])
                        )
                        enum_id_idx += 1
            if self.logger is not None:
                self.logger.log(
                    "Enforced IDs for compounds and enumerations specified (merging disabled).",
                    _LE.DEBUG,
                )
        return compounds



import os
import shutil
import tempfile
from icolos.core.containers.perturbation_map import Edge, Node, PerturbationMap
from icolos.core.workflow_steps.pmx.base import StepPMXBase
from pydantic import BaseModel
from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.enums.program_parameters import (
    GromacsEnum,
    PMXAtomMappingEnum,
    PMXEnum,
)
from icolos.utils.enums.step_enums import StepGromacsEnum
from icolos.utils.execute_external.execute import Executor

from icolos.utils.execute_external.gromacs import GromacsExecutor

_SGE = StepGromacsEnum()
_GE = GromacsEnum()
_PE = PMXEnum()
_PAE = PMXAtomMappingEnum()
_LE = LoggingConfigEnum()


class StepPMXmutate(StepPMXBase, BaseModel):
    """This is the de facto entrypoint for a protein FEP workflow

    This step does the legwork for setting up the output dir structure, similar to pmx_setup for the small molecule case
    """

    def __init__(self, **data):
        super().__init__(**data)
        self._initialize_backend(Executor)

    def execute(self):
        # start with a pdb file, parametrize with pdb2gmx
        replicas = (
            self.settings.additional["replicas"]
            if "replicas" in self.settings.additional.keys()
            else 3
        )
        assert self.work_dir is not None

        # this is a bit awkward putting them in the ligand dir, we adopt the same structure as the small molecule approach
        os.makedirs(os.path.join(self.work_dir, "input"), exist_ok=True)
        for folder in ["ligands", "mdp", "protein"]:
            os.makedirs(os.path.join(self.work_dir, "input", folder), exist_ok=True)
        setup_dir = os.path.join(self.work_dir, _PAE.LIGAND_DIR)
        self.data.generic.write_out_all_files(setup_dir)

        mdp_dir = self.data.generic.get_argument_by_extension(
            ext="mdp", rtn_file_object=True
        )
        mdp_dir.write(os.path.join(self.work_dir, "input/mdp"))

        # run pdb2gmx on the protein, generate a good input file
        pdb2gmx_args = [
            "-f",
            self.data.generic.get_argument_by_extension("pdb"),
            "-water",
            self.settings.additional["water"],
            "-ff",
            self.settings.additional["forcefield"],
            "-o",
            # use pdb here to retain chain IDs for pmx mutate
            os.path.join(setup_dir, "wt.pdb"),
            "-ignh",
        ]
        # run the
        self._gromacs_executor.execute(
            command=_GE.PDB2GMX, arguments=pdb2gmx_args, check=True, location=setup_dir
        )
        # now we have the prepared structure to pass to pmx mutate

        # parse mutant file, note that pdb2gmx will remove chain IDs
        with open(
            os.path.join(setup_dir, self.data.generic.get_argument_by_extension("mut")),
            "r",
        ) as f:
            mut_lines = [l for l in f.readlines() if l.strip()]

        perturbation_map = PerturbationMap()
        wt_node = Node(node_id="wt", node_hash="wt")
        perturbation_map.add_node(wt_node)
        # all we need is the identities of the mutants

        # in the form "P 76 TYR\n X 45 GLY"
        for mut in mut_lines:
            chain, resid, new_res = mut.split()
            edge_dir = os.path.join(self.work_dir, f"wt_{chain}{resid}{new_res}")
            os.makedirs(
                edge_dir,
                exist_ok=True,
            )
            for branch in self.therm_cycle_branches:
                for state in self.states:
                    for r in range(1, replicas + 1):
                        for sim in self.sim_types:
                            os.makedirs(
                                os.path.join(
                                    self.work_dir,
                                    edge_dir,
                                    branch,
                                    state,
                                    f"run{r}",
                                    sim,
                                ),
                                exist_ok=True,
                            )
            mutate_args = [
                "-f",
                os.path.join(setup_dir, "wt.pdb"),
                "--ref",
                os.path.join(setup_dir, "wt.pdb"),
                "-o",
                f"{chain}{resid}{new_res}.pdb",
                "-ff",
                self._get_additional_setting(_SGE.FORCEFIELD),
            ]
            self._logger.log(
                f"Generating mutation {chain}{resid}->{new_res}", _LE.DEBUG
            )
            result = self._backend_executor.execute(
                command=_PE.MUTATE,
                arguments=mutate_args,
                location=os.path.join(edge_dir, "bound"),
                check=True,
                pipe_input=f'echo -e "{chain}\n{resid}\n{new_res}\nn"',
            )

            # run pdb2gmx on the resulting structure, resulting structures are for the bound states

            pdb2gmx_args = [
                "-f",
                f"{chain}{resid}{new_res}.pdb",
                "-water",
                self.settings.additional["water"],
                "-ff",
                self.settings.additional["forcefield"],
                "-o",
                "init.pdb",
            ]
            self._backend_executor.execute(
                command=_GE.PDB2GMX,
                arguments=pdb2gmx_args,
                location=os.path.join(edge_dir, "bound"),
                check=True,
            )
            # now make the directory in the main workdir to  for that mutation and copy the input files over

            # now we have a single mutated structure file per system.
            node = Node(
                node_id=f"{chain}{resid}{new_res}",
                node_hash=f"{chain}{resid}{new_res}",
            )
            perturbation_map.add_node(node)
            perturbation_map.add_edge(Edge(node_from=wt_node, node_to=node))

            # generate connectivity (just a star map), now we have an initialized perturbation map, relying on file paths only, not attached structures for now
            for node in perturbation_map.nodes:
                perturbation_map._attach_node_connectivity(node)

            self.get_workflow_object().set_perturbation_map(perturbation_map)

            # TODO: we need to generate mutatinos for the bound state and the unbound state, by separating the pdb file

            # we have the chain being mutated, extract this to its own file

            # create the index file
            make_ndx_args = ["-f", "init.pdb"]
            self._backend_executor.execute(
                command=_GE.MAKE_NDX,
                arguments=make_ndx_args,
                location=os.path.join(edge_dir, "bound"),
                pipe_input=f'echo -e "chain {chain}\nq"',
                check=True,
            )
            # index.ndx in the bound state directory

            editconf_args = [
                "-f",
                "init.pdb",
                "-n",
                "index.ndx",
                "-o",
                os.path.join(
                    edge_dir, "unbound", f"{chain}{resid}{new_res}_unbound.pdb"
                ),
            ]
            self._backend_executor.execute(
                command=_GE.EDITCONF,
                arguments=editconf_args,
                location=os.path.join(edge_dir, "bound"),
                check=True,
                pipe_input=f'echo -e "ch{chain}\n"',
            )

            # generate topology for the unbound state
            pdb2gmx_args = [
                "-f",
                f"{chain}{resid}{new_res}_unbound.pdb",
                "-water",
                self.settings.additional["water"],
                "-ff",
                self.settings.additional["forcefield"],
                "-o",
                "init.pdb",
            ]
            self._backend_executor.execute(
                command=_GE.PDB2GMX,
                arguments=pdb2gmx_args,
                location=os.path.join(edge_dir, "unbound"),
                check=True,
            )


help_string = """
pmx mutate -h
usage: pmx [-h] [-f infile] [-fB infileB] [-o outfile] [-ff ff]                                    
           [--script script] [--keep_resid | --ref ] [--resinfo]                                   
                                                                                                   
This script applies mutations of residues in a structure file for subsequent                                          
free energy calculations. It supports mutations to protein, DNA, and RNA                           
molecules.                                                                                         
                                                                          
The mutation information and dummy placements are taken from the hybrid residue
database "mutres.mtp". The best way to use this script is to take a pdb/gro file                                                                    
that has been written with pdb2gmx with all hydrogen atoms present.
                                                                          
By default, all residues are renumbered starting from 1, so to have unique
residue IDs. If you want to keep the original residue IDs, you can use the flag                                                                     
--keep_resid. In this case, you will also need to provide chain information                                                                         
in order to be able to mutate the desired residue. Alternatively, if you would                                                                      
like to use the original residue IDs but these have been changed, e.g. by gromacs,
you can provide a reference PDB file (with chain information too) using the --ref
flag. The input structure will be mutated according to the IDs chosen for the
reference structure after having mapped the two residue indices.

The program can either be executed interactively or via script. The script file
simply has to consist of "residue_id target_residue_name" pairs (just with some
space between the id and the name), or "chain_id residue_id target_residue_name"
if you are keeping the original residue IDs or providing a reference structure.

The script uses an extended one-letter code for amino acids to account for
different protonation states. Use the --resinfo flag to print the dictionary.

optional arguments:
  -h, --help       show this help message and exit
  -f infile        Input structure file in PDB or GRO format. Default is "protein.pdb"
  -fB infileB      Input structure file of the B state in PDB or GRO format (optional).
  -o outfile       Output structure file in PDB or GRO format. Default is "mutant.pdb"
  -ff ff           Force field to use. If none is provided, 
                   a list of available ff will be shown.
  --script script  Text file with list of mutations (optional).
  --keep_resid     Whether to renumber all residues or to keep the
                   original residue IDs. By default, all residues are
                   renumbered so to have unique IDs. With this flags set,
                   the original IDs are kept. Because the IDs might not
                   be unique anymore, you will also be asked to choose
                   the chain ID where the residue you want to mutate is.
  --ref            Provide a reference PDB structure from which to map
                   the chain and residue IDs onto the file to be mutated (-f).
                   This can be useful when wanting to mutate a file that
                   has had its residues renumbered or the chain information
                   removed (e.g. after gmx grompp). As in the --keep_resid
                   option, if --ref is chosen, you will need to provide chain
                   information either interactively or via the --script flag.
  --resinfo        Show the list of 3-letter -> 1-letter residues
"""



from typing import List
from pydantic import BaseModel


class ConfigProteinData(BaseModel):
    protein_datafile: str


class ConfigFitnessFunctionSettings(BaseModel):
    initial_virtual_pt_match_max: int = 3
    relative_ligand_energy: int = 1
    gold_fitfunc_path: str = "goldscore"
    start_vdw_linear_cutoff: float = 6
    score_param_file: str = "DEFAULT"


class ConfigSaveOptions(BaseModel):
    save_score_in_file: int = 1
    save_protein_torsions: int = 1


class ConfigCovalentBonding(BaseModel):
    # TODO: extend to support covalent docking
    covalent: int = 0


class ConfigConstraints(BaseModel):
    # TODO: extend to support constrained docking
    force_constraints: int = 0


class ConfigTermination(BaseModel):
    early_termination: int = 1
    n_top_solutions: int = 3
    rms_tolerance: float = 1.5


class ConfigFlags(BaseModel):
    internal_ligand_h_bonds: int = 0
    flip_free_corners: int = 0
    match_ring_templates: int = 0
    flip_amide_bonds: int = 0
    flip_planar_n: str = "1 flip_ring_NRR flip_ring_NHR"
    flip_pyramidal_n: str = 0
    rotate_carboxylic_oh: str = "flip"
    use_tordist: int = 1
    postprocess_bonds: int = 1
    rotatable_bond_override_file: str = "DEFAULT"
    solvate_all: int = 1


class ConfigDataFiles(BaseModel):
    ligand_data_file: List[str] = []
    param_file: str = "DEFAULT"
    set_ligand_atom_types: int = 1
    set_protein_atom_types: int = 0
    directory: str = "."
    tordist_file: str = "DEFAULT"
    make_subdirs: int = 0
    save_lone_pairs: int = 1
    fit_points_file: str = "fit_pts.mol2"
    read_fitpts: str = 0


class ConfigFloodFill(BaseModel):
    radius: float = 10
    origin: str = "0 0 0"
    do_cavity: int = 1
    floodfill_atom_no: int = 0
    cavity_file: str
    floodfill_center: str = "cavity_from_ligand 7 atoms"


class ConfigGeneticOperators(BaseModel):
    pt_crosswt: str = "auto"
    allele_mutatewt: str = "auto"
    migratewt: str = "auto"


class ConfigPopulation(BaseModel):
    # Note that pydantic accepts also numbers when strings are specified (unless "strict" mode is utilized)
    popsiz: str = "auto"
    select_pressure: str = "auto"
    n_islands: str = "auto"
    maxops: str = "auto"
    niche_siz: str = "auto"


class ConfigAutomaticSettings(BaseModel):
    autoscale: float = (
        1.0  # between 0 (off) and 5.0 (the larger the slower but more sampling)
    )
    autoscale_nops_max: int = (
        100000  # maximum value for autoscale operations (0 means off)
    )
    autoscale_nops_min: int = (
        100  # minimum value for autoscale operations (0 means off)
    )



import gzip
import os
import shutil
import tempfile
from copy import deepcopy
from typing import List, Tuple

from pydantic import BaseModel
from rdkit import Chem

from icolos.core.workflow_steps.schrodinger.base import StepSchrodingerBase
from icolos.utils.execute_external.glide import GlideExecutor
from icolos.utils.execute_external.schrodinger import SchrodingerExecutor
from icolos.utils.general.files_paths import any_in_file, gen_tmp_file

from icolos.core.containers.compound import Conformer

from icolos.utils.enums.program_parameters import SchrodingerExecutablesEnum, GlideEnum
from icolos.utils.enums.step_enums import StepGlideEnum, StepBaseEnum
from icolos.core.workflow_steps.step import _LE
from icolos.utils.general.parallelization import SubtaskContainer, Parallelizer, Subtask
from icolos.utils.general.strings import stringify


class GlideSupportEnum:

    GLIDE_INPUTBLOCK_COMMASEPARATED = [
        "CONSTRAINT_GROUP"
    ]  # define list of block keys which are to have commas
    GLIDE_INPUTBLOCK_VALUEQUOTED = [
        "FEATURE"
    ]  # define list of block keys, where values are to be put
    # into double quotation marks

    GLIDE_TG_WAIT_INTERVAL = "wait_interval_seconds"
    GLIDE_TG_WAIT_LIMIT = "wait_limit_seconds"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


_SBE = StepBaseEnum
_EE = GlideEnum()
_SGE = StepGlideEnum()
_SEE = SchrodingerExecutablesEnum()
_GSE = GlideSupportEnum()


class StepGlide(StepSchrodingerBase, BaseModel):

    _schrodinger_executor: SchrodingerExecutor = None

    class Config:
        underscore_attrs_are_private = True

    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executors and test availability
        self._initialize_backend(executor=GlideExecutor)
        self._check_backend_availability()

        self._schrodinger_executor = SchrodingerExecutor(
            prefix_execution=self.execution.prefix_execution,
            binary_location=self.execution.binary_location,
        )

    def _get_scores_from_conformer(self, conformer: Chem.Mol) -> Tuple[float, float]:
        return (
            float(conformer.GetProp(_SGE.GLIDE_DOCKING_SCORE)),
            float(conformer.GetProp(_SGE.GLIDE_GSCORE)),
        )

    def _set_docking_score(self, conformer: Chem.Mol) -> bool:
        try:
            docking_score, g_score = self._get_scores_from_conformer(conformer)
        except KeyError:
            return False
        conformer.SetProp(_SBE.ANNOTATION_TAG_DOCKING_SCORE, str(docking_score))
        conformer.SetProp(_SBE.ANNOTATION_TAG_G_SCORE, str(g_score))
        return True

    def _generate_temporary_input_output_files(
        self, batch: List[List[Subtask]]
    ) -> Tuple[List[str], List[str], List[str], List[str]]:
        tmp_output_dirs = []
        tmp_input_mae_paths = []
        tmp_output_sdf_paths = []
        tmp_output_maegz_paths = []

        for next_subtask_list in batch:
            # generate temporary input files and output directory
            cur_tmp_output_dir = tempfile.mkdtemp()
            _, cur_tmp_sdf = gen_tmp_file(suffix=".sdf", dir=cur_tmp_output_dir)
            _, cur_tmp_mae = gen_tmp_file(suffix=".mae", dir=cur_tmp_output_dir)

            # write-out the temporary input file
            writer = Chem.SDWriter(cur_tmp_sdf)
            one_written = False
            for subtask in next_subtask_list:
                enumeration = subtask.data
                mol = deepcopy(enumeration.get_molecule())
                if mol is not None:
                    mol.SetProp("_Name", enumeration.get_index_string())
                    one_written = True
                    writer.write(mol)
            writer.close()
            if one_written is False:
                self._remove_temporary(cur_tmp_output_dir)
                continue

            # translate the SDF into a MAE file
            self._translate_SDF_to_MAE(
                sdf_path=cur_tmp_sdf,
                mae_path=cur_tmp_mae,
                executor=self._schrodinger_executor,
            )

            # add the path to which "_dock_subjob()" will write the result SDF
            _, output_sdf_path = gen_tmp_file(
                suffix="_result.sdf", dir=cur_tmp_output_dir
            )
            _, output_maegz_path = gen_tmp_file(
                suffix="_result.maegz", dir=cur_tmp_output_dir, text=False
            )
            tmp_output_sdf_paths.append(output_sdf_path)
            tmp_output_maegz_paths.append(output_maegz_path)
            tmp_input_mae_paths.append(cur_tmp_mae)
            tmp_output_dirs.append(cur_tmp_output_dir)
        return (
            tmp_output_dirs,
            tmp_input_mae_paths,
            tmp_output_sdf_paths,
            tmp_output_maegz_paths,
        )

    def _all_keywords(self) -> dict:
        """Returns joined keywords from JSON and from .in file (if specified)."""

        keywords = {}

        # keywords from maestro file; they can be overwritten by explicitly set values from the "configuration" block
        maestro_in_file = deepcopy(
            self.settings.additional.get(_SGE.MAESTRO_IN_FILE, None)
        )
        if maestro_in_file is not None:
            with open(maestro_in_file[_SGE.MAESTRO_IN_FILE_PATH], "rt") as f:
                keywords_from_file = self._parse_maestro_in_file(f.readlines())
                keywords.update(keywords_from_file)

        # Add keywords from advanced_glide_keywords
        # (they are keywords with file paths),
        # skipping keywords that are None.
        # Also skip maestro file - that's not a keyword.
        # TODO: This is legacy code from DockStream's implementation, which was necessary to accommodate the GUI.
        #       Remove?
        # if self.parameters.advanced_glide_keywords is not None:
        #    adv_kw = stringify({
        #        k: v
        #        for k, v in self.parameters.advanced_glide_keywords.dict().items()
        #        if v is not None and k not in {'maestro_file'}
        #    })
        #    keywords.update(adv_kw)

        # Add "ordinary" keywords, overwriting existing ones.
        json_keywords = stringify(
            deepcopy(self.settings.additional.get(_SGE.CONFIGURATION, {}))
        )
        keywords.update(
            json_keywords
        )  # Overwrites any keywords that are already present.
        return keywords

    def _configuration_Maestro_reformat(self, configuration: dict):
        # rewrite keyword input file in Maestro format
        maestro_indent = "    "
        maestro_spacing = "   "

        element_lines = []
        block_lines = []

        for key in configuration.keys():
            if isinstance(configuration[key], str):
                # keyword holds one dictionary (string) only
                element_lines.append(
                    maestro_spacing.join([key, configuration[key] + "\n"])
                )
            elif isinstance(configuration[key], dict):
                # keyword holds a composite block and has no dictionary (e.g. constraints); note, that these must
                # always be at the end of the file
                block_lines.append("\n" + key + "\n")
                block = configuration[key]
                for key_idx, block_key in enumerate(block.keys()):
                    block_value = block[block_key]

                    # if this is a value in certain blocks, put it into double quotation marks as spaces are present
                    if any([x in key for x in _GSE.GLIDE_INPUTBLOCK_VALUEQUOTED]):
                        block_value = '"' + block_value + '"'
                    line = maestro_indent + maestro_spacing.join(
                        [block_key, block_value]
                    )

                    # add comma to block definition, if there are more lines to come and the block requires it
                    # note, that not all blocks in GLIDE require this; in some cases, the comma is already part of
                    # the line (then skip it!)
                    if any([x in key for x in _GSE.GLIDE_INPUTBLOCK_COMMASEPARATED]):
                        if (key_idx + 1) < len(block) and line[-1] != ",":
                            line = line + ","

                    block_lines.append(line + "\n")
            else:
                raise Exception(
                    f"Cannot handle type {type(configuration[key])} in configuration file specification, only use strings and blocks."
                )

        return element_lines, block_lines

    def _write_configuration_to_file(self, configuration: dict, path: str):
        """Function to generate a keyword input file in Maestro format."""

        # call a function that returns the input keywords in Maestro format
        element_lines, block_lines = self._configuration_Maestro_reformat(
            configuration=configuration
        )

        # arrange the elements and blocks
        if path is None:
            _, path = gen_tmp_file(suffix=".in")
        with open(path, mode="w") as f:
            self._logger.log(f"Writing GLIDE input file {path}:\n", _LE.DEBUG)
            for line in element_lines:
                f.write(line)
                self._logger_blank.log(line.rstrip("\n"), _LE.DEBUG)
            for line in block_lines:
                f.write(line)
                self._logger_blank.log(line.rstrip("\n"), _LE.DEBUG)
            self._logger_blank.log("", _LE.DEBUG)
            self._logger.log("--- End file", _LE.DEBUG)

    def _get_time_limit_per_task(self):
        # for "SP" method, it can be expected to that about 90 s / ligand is required at most
        # use a bit extra
        return int(self.settings.additional.get(_SGE.TIME_LIMIT_PER_TASK, 240))

    def _get_path_tmp_results(
        self, glide_pose_outtype: str, base_path: str
    ) -> Tuple[str, str]:
        if glide_pose_outtype == _EE.GLIDE_POSE_OUTTYPE_LIGANDLIB:
            path_tmp_results = os.path.join(
                os.path.dirname(base_path),
                "".join(
                    [
                        os.path.splitext(os.path.basename(base_path))[0],
                        _SGE.GLIDE_SDF_DEFAULT_EXTENSION,
                    ]
                ),
            )
        elif glide_pose_outtype == _EE.GLIDE_POSE_OUTTYPE_POSEVIEWER:
            path_tmp_results = os.path.join(
                os.path.dirname(base_path),
                "".join(
                    [
                        os.path.splitext(os.path.basename(base_path))[0],
                        _SGE.GLIDE_MAEGZ_DEFAULT_EXTENSION,
                    ]
                ),
            )
        else:
            raise NotImplementedError(
                f"Specified out-type {glide_pose_outtype} for Glide not supported."
            )

        path_tmp_log = os.path.join(
            os.path.dirname(base_path),
            "".join([os.path.splitext(os.path.basename(base_path))[0], _SGE.GLIDE_LOG]),
        )
        return path_tmp_results, path_tmp_log

    def _move_result_files(
        self,
        glide_pose_outtype: str,
        path_tmp_results: str,
        path_sdf_results: str,
        path_maegz_results: str,
    ):
        if glide_pose_outtype == _EE.GLIDE_POSE_OUTTYPE_LIGANDLIB:
            if os.path.isfile(path_tmp_results):
                with gzip.open(path_tmp_results, "rb") as fin:
                    with open(path_sdf_results, "wb") as fout:
                        shutil.copyfileobj(fin, fout)
        elif glide_pose_outtype == _EE.GLIDE_POSE_OUTTYPE_POSEVIEWER:
            # as the output is in MAEGZ format, we need to translate it into an SDF (and move the original file to the
            # expected path)
            self._translate_MAE_to_SDF(
                mae_path=path_tmp_results,
                sdf_path=path_sdf_results,
                executor=self._schrodinger_executor,
            )
            os.rename(path_tmp_results, path_maegz_results)
        else:
            raise NotImplementedError(
                f"Specified out-type {glide_pose_outtype} for Glide not supported."
            )

    def _run_subjob(
        self,
        mae_ligand_path,
        path_sdf_results,
        path_maegz_results,
        tmp_output_dir,
        grid_path,
        sublist,
    ):
        # 1) increase the sublist "tries" and set status to "failed"
        _ = [task.increment_tries() for task in sublist]
        _ = [task.set_status_failed() for task in sublist]

        # 2) change to directory, to be able to use relative paths (to compensate for Schrodinger bug with AWS)
        working_dir = os.getcwd()
        os.chdir(tmp_output_dir)

        # 3) get "keywords" dictionary and overwrite necessary values
        #    add "LIGANDFILE" keyword to list of keywords: full path to "mae" formatted ligands
        configuration = self._all_keywords()
        if configuration is None:
            raise ValueError(
                f"You need to specify at least the gridfile path in the configuration for Glide."
            )
        configuration[_EE.GLIDE_LIGANDFILE] = mae_ligand_path

        # set the path to the grid file for this run
        configuration[_EE.GLIDE_GRIDFILE] = grid_path

        # if not set, set the liand pose outtype to "LIGANDLIB" (SDF output without receptor)
        glide_pose_outtype = configuration.get(
            _EE.GLIDE_POSE_OUTTYPE, _EE.GLIDE_POSE_OUTTYPE_LIGANDLIB
        )
        configuration[_EE.GLIDE_POSE_OUTTYPE] = glide_pose_outtype

        # 4) write the keyword-input file for the "Glide" backend; write-out to temporary file
        _, glide_configuration_path = gen_tmp_file(suffix=".in", dir=tmp_output_dir)
        self._write_configuration_to_file(
            configuration=configuration,
            path=glide_configuration_path,
        )

        # 5) wait / sleep until job is completed
        #    Note, that while Glide has an option "-WAIT", this does not seem to work when getting back
        #    data from AWS (probably it ends before copying back the data properly); stay with this solution for now
        path_tmp_results, path_tmp_log = self._get_path_tmp_results(
            glide_pose_outtype=glide_pose_outtype, base_path=glide_configuration_path
        )

        # 5) execute the "Glide" backend
        arguments = self._prepare_glide_arguments(glide_configuration_path)
        execution_result = self._backend_executor.execute(
            command=_EE.GLIDE,
            arguments=arguments,
            check=True,
            location=os.path.dirname(glide_configuration_path),
        )

        # 6) check return code (anything but '0' is bad) and add "stdout" to log file
        time_exceeded = False
        if execution_result.returncode != 0:
            msg = (
                f"Could not dock with Glide, error message: {execution_result.stdout}."
            )
            self._logger.log(msg, _LE.ERROR)
            self._print_log_file(path_tmp_log)
            raise RuntimeError()
        else:
            if (
                self._wait_until_file_generation(
                    path=path_tmp_results,
                    path_log=path_tmp_log,
                    interval_sec=10,
                    maximum_sec=max(
                        self._get_time_limit_per_task() * len(sublist), 300
                    ),
                    success_strings=_EE.GLIDE_LOG_FINISHED_STRINGS,
                    fail_strings=_EE.GLIDE_LOG_FAIL_STRINGS,
                )
                is False
            ):
                time_exceeded = True
                self._logger.log(
                    f"Sublist docking for output file {path_tmp_results} exceeded time limit or failed, "
                    f"all these ligands are ignored in the final write-out. This could mean that none of "
                    f"them could be docked or a runtime error in Glide occured.",
                    _LE.DEBUG,
                )

        # 6) load the log-file (if generated) and check if all went well
        if (
            any_in_file(path_tmp_log, _EE.GLIDE_LOG_SUCCESS_STRING)
            and time_exceeded is False
        ):
            self._logger.log(
                f"Finished sublist (input: {mae_ligand_path}, output: {path_sdf_results}).",
                _LE.DEBUG,
            )
        else:
            self._print_log_file(path_tmp_log)

        # 7) collect the results; Glide outputs the sdf with a given, semi-hard-coded path; extract the sdf file
        self._move_result_files(
            glide_pose_outtype=glide_pose_outtype,
            path_tmp_results=path_tmp_results,
            path_sdf_results=path_sdf_results,
            path_maegz_results=path_maegz_results,
        )

        # 8) revert back to working directory
        os.chdir(working_dir)

    def _prepare_glide_arguments(self, glide_configuration_path: str) -> List[str]:
        # Note, that the first argument is the path to the configuration input file
        # If the number of cores has been set, overwrite "N_JOBS" and parallelize internally and also note
        # that each subjob requires a license; instead start each with "N_JOBS" = 1
        arguments = [glide_configuration_path]

        # copy parameters and overwrite as necessary
        parameters = deepcopy(self.settings.arguments.parameters)
        parameters[_EE.GLIDE_NJOBS] = 1

        if len(self.settings.arguments.flags) > 0:
            for flag in self.settings.arguments.flags:
                # -WAIT leads to issues at times: The process may not return properly
                # (e.g. because of writing problems) and then gets stuck; workaround with waiting
                # for file completion, so remove it if set
                if flag not in [_EE.GLIDE_WAIT]:
                    arguments.append(str(flag))
        if parameters:
            for key in parameters.keys():
                # remove "-WAIT" if set as a parameter, as this leads to instability issues and ignore empty keys
                if key == _EE.GLIDE_WAIT or key == "":
                    continue
                arguments.append(key)
                if parameters[key] is not None and parameters[key] != "":
                    arguments.append(str(parameters[key]))
        return arguments

    def _execute_glide(self, grid_id: str, grid_path: str):
        # TODO: add individual resubmission for failed subtasks
        # get number of sublists in batch and initialize Parallelizer
        glide_parallelizer = Parallelizer(func=self._run_subjob)

        # continue until everything is successfully done or number of retries have been exceeded
        while self._subtask_container.done() is False:
            next_batch = self._get_sublists(get_first_n_lists=self._get_number_cores())

            # generate paths and initialize molecules (so that if they fail, this can be covered)
            (
                tmp_output_dirs,
                tmp_input_mae_paths,
                tmp_output_sdf_paths,
                tmp_output_maegz_paths,
            ) = self._generate_temporary_input_output_files(next_batch)

            # call "token guard" method (only executed, if block is specified in the configuration), which will wait
            # with the execution if not enough tokens are available at the moment
            self._apply_token_guard()

            # execute the current batch in parallel; hand over lists of parameters (will be handled by Parallelizer)
            # also increment the tries and set the status to "failed" (don't do that inside subprocess, as data is
            # copied, not shared!)
            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]
            list_grid_path = [grid_path for _ in tmp_input_mae_paths]
            glide_parallelizer.execute_parallel(
                mae_ligand_path=tmp_input_mae_paths,
                path_sdf_results=tmp_output_sdf_paths,
                path_maegz_results=tmp_output_maegz_paths,
                tmp_output_dir=tmp_output_dirs,
                grid_path=list_grid_path,
                sublist=next_batch,
            )

            # parse the output of that particular batch and remove temporary files
            self._parse_glide_output(
                tmp_output_sdf_paths,
                tmp_output_maegz_paths,
                next_batch,
                grid_id,
                grid_path,
            )

            # clean-up
            self._remove_temporary(tmp_output_dirs)

            # print the progress for this execution
            self._log_execution_progress()

    def _log_execution(self, grid_id: str, number_grids: int):
        number_enumerations = 0
        number_conformers = 0
        for compound in self.get_compounds():
            number_enumerations += len(compound)
            for enumeration in compound:
                number_conformers += len(enumeration)
                if len(enumeration) == 0:
                    self._logger.log(
                        f"Enumeration {enumeration.get_index_string()} has no docked poses attached.",
                        _LE.DEBUG,
                    )
        self._logger.log(
            f"Executed Schrodinger/Glide backend for grid {grid_id} (of {number_grids}), now storing a total of {number_conformers} conformers for {number_enumerations} enumerations in {len(self.get_compounds())} compounds.",
            _LE.INFO,
        )

    def _parse_glide_output(
        self,
        tmp_output_sdf_paths: List[str],
        tmp_output_maegz_paths: List[str],
        batch: List[List[Subtask]],
        grid_id: str,
        grid_path: str,
    ):
        # TODO: refactor that (recombine with ligprep parsing?)
        def _update_subtask(sublist: List[Subtask], enum_identifier: str):
            for task in sublist:
                if task.data.get_index_string() == enum_identifier:
                    task.set_status_success()

        def _add_poseviewer_file(conformer: Conformer, maegz_path: str):
            if os.path.isfile(maegz_path) and os.path.getsize(maegz_path) > 0:
                with open(maegz_path, "rb") as f:
                    conformer.add_extra_data(
                        key=_SGE.GLIDE_POSEVIEWER_FILE_KEY, data=f.read()
                    )

        for i in range(len(tmp_output_sdf_paths)):
            # get input and output paths and check the files are there
            path_sdf_results = tmp_output_sdf_paths[i]
            path_maegz_results = tmp_output_maegz_paths[i]
            cur_sublist = batch[i]

            # this is a protection against the case where empty (file size == 0 bytes) files are generated due to
            # a failure during docking
            if (
                not os.path.isfile(path_sdf_results)
                or os.path.getsize(path_sdf_results) == 0
            ):
                continue

            mol_supplier = Chem.SDMolSupplier(path_sdf_results, removeHs=False)
            for mol in mol_supplier:
                if mol is None:
                    continue
                cur_enumeration_name = str(mol.GetProp("_Name"))

                # add the information on the actual grid used
                mol.SetProp(_SBE.ANNOTATION_GRID_ID, str(grid_id))
                mol.SetProp(_SBE.ANNOTATION_GRID_PATH, str(grid_path))
                mol.SetProp(_SBE.ANNOTATION_GRID_FILENAME, os.path.basename(grid_path))

                # if no docking score is attached (i.e. the molecule is a receptor or so, skip it)
                if self._set_docking_score(mol) is not True:
                    continue

                # add molecule to the appropriate ligand
                for compound in self.get_compounds():
                    for enumeration in compound:
                        if enumeration.get_index_string() == cur_enumeration_name:
                            new_conformer = Conformer(
                                conformer=mol,
                                conformer_id=None,
                                enumeration_object=enumeration,
                            )
                            _add_poseviewer_file(
                                conformer=new_conformer, maegz_path=path_maegz_results
                            )
                            enumeration.add_conformer(new_conformer, auto_update=True)
                            _update_subtask(
                                cur_sublist, enum_identifier=cur_enumeration_name
                            )
                            break

    def _fill_dummy_conformers(self):
        # if any enumerations have not got poses attached by this point, docking failed.  Attach a dummy conformer with the same mol object as the enumeration with gscore = 0
        num_added = 0
        for compound in self.get_compounds():
            for enum in compound.get_enumerations():
                if not enum.get_conformers():
                    num_added += 1
                    self._logger.log(
                        f"Added dummy conformer for enumeration {enum.get_enumeration_id()}!",
                        _LE.DEBUG,
                    )
                    dummy_conf = Conformer(
                        conformer=enum.get_molecule(), conformer_id=0
                    )
                    dummy_conf.get_molecule().SetProp(
                        _SBE.ANNOTATION_TAG_DOCKING_SCORE, str(0.0)
                    )
                    dummy_conf.get_molecule().SetProp(
                        _SGE.GLIDE_DOCKING_SCORE, str(0.0)
                    )
                    enum.add_conformer(dummy_conf)
        self._logger.log(f"Added {num_added} dummy conformers", _LE.DEBUG)

    def _sort_conformers(self):
        # sort the conformers (best to worst) and update their names to contain the conformer id
        # -> <compound>:<enumeration>:<conformer_number>
        for compound in self.get_compounds():
            for enumeration in compound:
                enumeration.sort_conformers(
                    by_tag=_SGE.GLIDE_DOCKING_SCORE, reverse=False
                )

    def execute(self):
        # in order to be able to efficiently execute Glide on the enumeration level, all of them have to be unrolled
        # Note: As they retain their respective Compound object, the attribution later on is simple
        all_enumerations = []
        for compound in self.get_compounds():
            all_enumerations = all_enumerations + compound.get_enumerations()
            for enumeration in compound:
                enumeration.clear_conformers()

        # to allow ensemble docking, loop over all provided grid files and annotate the origin of the conformers
        gridfiles = deepcopy(self.settings.additional.get(_SGE.CONFIGURATION, None))[
            _EE.GLIDE_GRIDFILE
        ]
        if not isinstance(gridfiles, list):
            gridfiles = [gridfiles]

        # set grid ids (generate indices, if not specified)
        grid_ids = self.settings.additional.get(_SBE.GRID_IDS, [])
        if len(grid_ids) != len(gridfiles):
            self._logger.log(
                f"There were {len(grid_ids)} grid_ids specified for {len(gridfiles)}, using indices instead.",
                _LE.DEBUG,
            )
            grid_ids = [str(idx) for idx in range(len(gridfiles))]

        for grid_id, grid_path in zip(grid_ids, gridfiles):
            # split into sublists, according to the settings
            self._subtask_container = SubtaskContainer(
                max_tries=self.execution.failure_policy.n_tries
            )
            self._subtask_container.load_data(all_enumerations)

            # execute Glide
            self._execute_glide(grid_id=grid_id, grid_path=grid_path)

            # do the logging
            self._log_execution(grid_id=grid_id, number_grids=len(gridfiles))

        if self._get_additional_setting(_SGE.FILL_DUMMY_CONFS, default=False):
            self._fill_dummy_conformers()

        # sort the conformers loaded to the enumerations
        self._sort_conformers()



from icolos.core.workflow_steps.pmx.base import StepPMXBase
from pydantic import BaseModel


class StepPMXdoublebox(StepPMXBase, BaseModel):
    """Place two input structures into a single box."""

    def __init__(self, **data):
        super().__init__(**data)

    def execute(self):
        pass


help_string = """
pmx doublebox -h
usage: pmx [-h] -f1  -f2  [-o] [-r] [-d] [--longest_axis]

Places two structures into a single box. The box is a rectangular cuboid in
which the two structures are placed in such a way as to minimise the box
volume. You can use this script to help in the setup of a calculation using
the single-box double-system approach.

optional arguments:
  -h, --help      show this help message and exit
  -f1             First structure in PDB or GRO format.
  -f2             Second structure in PDB or GRO format.
  -o              Name of output file. Default is "doublebox.pdb".
  -r              Distance between the two structures (nm). Default is 2.5 nm.
  -d              Distance to the box wall (nm). Default is 1.5 nm.
  --longest_axis  Whether to just place structures along the longest axis,
                  rather then minimising the volume. Default is False.
"""



from icolos.core.workflow_steps.schrodinger.prime import StepPrime
from icolos.core.workflow_steps.schrodinger.macromodel import StepMacromodel
from icolos.core.workflow_steps.schrodinger.ligprep import StepLigprep
from icolos.core.workflow_steps.schrodinger.glide import StepGlide
from icolos.core.workflow_steps.schrodinger.prepwizard import StepPrepwizard
from icolos.core.workflow_steps.schrodinger.fep_plus_setup import StepFepPlusSetup
from icolos.core.workflow_steps.schrodinger.fep_plus_execution import StepFepPlusExec
from icolos.core.workflow_steps.schrodinger.desmond_preprocessor import StepDesmondSetup
from icolos.core.workflow_steps.schrodinger.desmond_exec import StepDesmondExec
from icolos.core.workflow_steps.schrodinger.residue_scanning import StepResidueScanning



class InputEnum:

    SOURCE_FIELD_COMPOUNDS = "compounds"
    TARGET_FIELD_COMPOUNDS = "compounds"
    TARGET_FIELD_CONFORMERS = "conformers"

    # REINVENT-compatible JSON input
    JSON_NAMES = "names"
    JSON_SMILES = "smiles"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")



from icolos.utils.enums.step_enums import StepBaseEnum, StepGromacsEnum
from icolos.utils.enums.program_parameters import GromacsEnum
from icolos.core.workflow_steps.gromacs.base import StepGromacsBase
from icolos.utils.execute_external.gromacs import GromacsExecutor
from pydantic import BaseModel
from icolos.core.workflow_steps.step import _LE
import os

_GE = GromacsEnum()
_SGE = StepGromacsEnum()
_SBE = StepBaseEnum


class StepGMXGenion(StepGromacsBase, BaseModel):
    """
    Wrapper for gmx genion
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=GromacsExecutor)
        self._check_backend_availability()

    def execute(self):
        tmp_dir = self._make_tmpdir()
        topol = self.get_topol()
        self.write_input_files(tmp_dir, topol=topol)
        arguments = self._parse_arguments(
            {
                "-o": _SGE.STD_STRUCTURE,
                "-p": _SGE.STD_TOPOL,
                "-s": _SGE.STD_TPR,
            }
        )
        result = self._backend_executor.execute(
            command=_GE.GENION,
            arguments=arguments,
            location=tmp_dir,
            pipe_input=self.construct_pipe_arguments(
                tmp_dir, self.settings.additional[_SBE.PIPE_INPUT]
            ),
        )
        for line in result.stdout.split("\n"):
            self._logger_blank.log(line, _LE.DEBUG)
        self._logger.log(
            f"Completed execution for {self.step_id} successfully", _LE.INFO
        )
        # this is the last structural change to the topology in a regular gromacs setup,
        # update the index groups here
        make_ndx_args = ["-f", _SGE.STD_STRUCTURE, "-o", _SGE.STD_INDEX]
        index_files = [f for f in os.listdir(tmp_dir) if f.endswith(".ndx")]
        # remove any existing index files
        for f in index_files:
            self._remove_temporary(os.path.join(tmp_dir, f))
        # generate new index file
        result = self._backend_executor.execute(
            command=_GE.MAKE_NDX,
            arguments=make_ndx_args,
            location=tmp_dir,
            check=True,
            pipe_input='echo -e "1 | 12 \nq"',
        )
        for line in result.stdout.split("\n"):
            self._logger_blank.log(line, _LE.DEBUG)

        self._logger.log('Added index group to "index.ndx"', _LE.DEBUG)
        self._parse_output(tmp_dir)
        topol.set_structure(tmp_dir)
        topol.parse(tmp_dir)
        self._remove_temporary(tmp_dir)



from typing import List, Union
from icolos.core.containers.compound import Compound
from icolos.core.containers.perturbation_map import Edge
from icolos.core.workflow_steps.pmx.base import StepPMXBase
from pydantic import BaseModel
from icolos.utils.enums.program_parameters import (
    StepPMXEnum,
)
from icolos.utils.execute_external.gromacs import GromacsExecutor
from icolos.utils.general.parallelization import SubtaskContainer
import os

_PSE = StepPMXEnum()


class StepPMXPrepareSimulations(StepPMXBase, BaseModel):
    """
    Prepare the tpr file for either equilibration or production simulations
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=GromacsExecutor)

    def execute(self):
        if self.run_type == _PSE.RBFE:
            edges = [e.get_edge_id() for e in self.get_edges()]
        elif self.run_type == _PSE.ABFE:
            edges = [c.get_index_string() for c in self.get_compounds()]
        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(edges)
        self._execute_pmx_step_parallel(
            run_func=self.prepare_simulation,
            step_id="pmx prepare_simulations",
            result_checker=self._check_result,
        )

    def prepare_simulation(self, jobs: List[Union[Edge, Compound]]) -> None:
        # define some constants that depend on whether this is rbfe/abfe
        # for abfe, edge refers to the ligand index
        sim_type = self.settings.additional[_PSE.SIM_TYPE]
        # FIXME: how do we get the replicas for abfe jobs without requiring input every time? inspect the workdir?
        replicas = (
            self.get_perturbation_map().replicas
            if self.get_perturbation_map() is not None
            else 1
        )

        for edge in jobs:
            for state in self.states:
                for r in range(1, replicas + 1):
                    for wp in self.therm_cycle_branches:
                        toppath = self._get_specific_path(
                            workPath=self.work_dir, edge=edge, wp=wp
                        )
                        # dir for the current sim type
                        simpath = self._get_specific_path(
                            workPath=self.work_dir,
                            edge=edge,
                            wp=wp,
                            state=state,
                            r=r,
                            sim=sim_type,
                        )
                        # dir for the previous sim type, from which we get confout.gro
                        prev_type = self._get_previous_sim_type(sim_type)
                        empath = self._get_specific_path(
                            workPath=self.work_dir,
                            edge=edge,
                            wp=wp,
                            state=state,
                            r=r,
                            sim=prev_type,
                        )
                        self._prepare_single_tpr(
                            simpath=simpath,
                            toppath=toppath,
                            state=state,
                            sim_type=sim_type,
                            empath=empath,
                            executor=self._backend_executor,
                        )

    def _get_previous_sim_type(self, sim_type: str):
        """
        Works out where to get starting structure from based on the current run and simulation type
        """
        if self._get_additional_setting(_PSE.PREV_STEP) is not None:
            return self._get_additional_setting(_PSE.PREV_STEP)
        elif self.run_type == _PSE.RBFE:
            if sim_type == "nvt":
                return "em"
            elif sim_type == "eq":
                return "nvt"
        elif self.run_type == _PSE.ABFE:
            if sim_type in ("em", "nvt"):
                return "em"
            elif sim_type == "npt":
                return "nvt"
            elif sim_type == "eq":
                return "npt"

    def _check_result(self, batch: List[List[str]]) -> List[List[bool]]:
        """
        Look in each hybridStrTop dir and check the output pdb files exist for the edges
        """
        sim_type = self.settings.additional[_PSE.SIM_TYPE]
        replicas = self.get_perturbation_map().replicas
        output_files = []
        for i in range(1, replicas + 1):
            output_files.append(f"unbound/stateA/run{i}/{sim_type}/tpr.tpr")
            output_files.append(f"unbound/stateB/run{i}/{sim_type}/tpr.tpr")
            output_files.append(f"bound/stateA/run{i}/{sim_type}/tpr.tpr")
            output_files.append(f"bound/stateB/run{i}/{sim_type}/tpr.tpr")
        results = []
        for subjob in batch:
            subjob_results = []
            for job in subjob:
                subjob_results.append(
                    all(
                        [
                            os.path.isfile(os.path.join(self.work_dir, job, f))
                            for f in output_files
                        ]
                    )
                )
            results.append(subjob_results)
        return results



import os
import shutil
import tempfile

from icolos.utils.enums.program_parameters import TurbomoleEnum
from icolos.utils.execute_external.execute import ExecutorBase

EE = TurbomoleEnum()


class TurbomoleExecutor(ExecutorBase):
    """For the execution of the "turbomole" binary."""

    def __init__(self, prefix_execution=None, binary_location=None):
        super().__init__(
            prefix_execution=prefix_execution, binary_location=binary_location
        )

    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):
        # check, whether a proper executable is provided
        if command not in [
            EE.TM_COSMOPREP,
            EE.TM_DEFINE,
            EE.TM_RIDFT,
            EE.TM_X2T,
            EE.TM_T2X,
            EE.CT_COSMOTHERM,
            EE.TM_JOBEX,
        ]:
            raise ValueError(
                "Parameter command must be an dictionary of the internal Turbomole executable list."
            )

        # TM accesses a folder specified in $TURBOTMPDIR to deposit the particular temporary files for a run; this is
        # system-wide, so parallel runs will interfere; also it is not removed automatically
        # TODO: find a more elegant solution; is this really necessary for all binaries or only "ridft" and "jobex"?
        tmp_dir = tempfile.mkdtemp()
        command = "".join(["export TURBOTMPDIR=", tmp_dir, " && ", command])

        result = super().execute(
            command=command,
            arguments=arguments,
            check=check,
            location=location,
            pipe_input=pipe_input,
        )

        if tmp_dir is not None and os.path.isdir(tmp_dir):
            shutil.rmtree(tmp_dir)
        return result

    def is_available(self):
        try:
            result = self.execute(command=EE.TM_RIDFT, arguments=[], check=True)

            if EE.TM_RIDFT_FAIL_IDENTIFICATION_STRING in result.stderr:
                return True
            return False
        except Exception as e:
            return False



import os
from icolos.loggers.blank_logger import BlankLogger


def print_log_file(path: str, logger, level):
    logger_blank = BlankLogger()
    if os.path.isfile(path):
        with open(path, "r") as log_file:
            log_file_raw = log_file.readlines()
            logger.log(f"Printing log file {path}:\n", level)
            for line in log_file_raw:
                logger_blank.log(line.rstrip("\n"), level)
            logger_blank.log("", level)
            logger.log("--- End file", level)



from pydantic import BaseModel
from rdkit import Chem

from icolos.utils.enums.program_parameters import OpenBabelEnum
from icolos.utils.enums.step_enums import StepAutoDockVinaTargetPreparationEnum
from icolos.utils.execute_external.autodockvina import AutoDockVinaExecutor
from icolos.utils.execute_external.openbabel import OpenBabelExecutor
from icolos.utils.general.icolos_exceptions import StepFailed

from icolos.core.workflow_steps.step import _LE, StepBase

_STE = StepAutoDockVinaTargetPreparationEnum()
_OBE = OpenBabelEnum()


class ADVExtractBoxTP(BaseModel):
    reference_ligand_path: str = None
    reference_ligand_format: str = _STE.EXTRACT_BOX_REFERENCE_LIGAND_FORMAT_PDB


class ADVAdditionalTP(BaseModel):
    pH: float = (
        7.4  # set target pH value that determines the protein's side-chain states
    )
    input_receptor_pdb: str = None
    output_receptor_pdbqt: str = None
    extract_box: ADVExtractBoxTP = ADVExtractBoxTP()


class StepAutoDockVinaTargetPreparation(StepBase, BaseModel):
    _openbabel_executor: OpenBabelExecutor = None
    adv_additional: ADVAdditionalTP = None

    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executor and test availability
        self._initialize_backend(executor=AutoDockVinaExecutor)
        self._check_backend_availability()

        # initialize the executor for all "OpenBabel"
        self._openbabel_executor = OpenBabelExecutor()
        if not self._openbabel_executor.is_available():
            raise StepFailed(
                "AutoDock Vina requires OpenBabel execution, initialization failed."
            )

        # set ADV specific settings and ensure that each molecule gets its own sublist
        self.adv_additional = ADVAdditionalTP(**self.settings.additional)

    def _export_as_pdb2pdbqt(self):
        # Note: In contrast to the ligand preparation, we will not use a tree-based flexibility treatment here - thus,
        #       the option "-xr" is used. Partial charges of the receptor are not used in AutoDock Vina.
        arguments = [
            " ".join(
                [_OBE.OBABEL_INPUTFORMAT_PDB, self.adv_additional.input_receptor_pdb]
            ),
            _OBE.OBABEL_OUTPUT_FORMAT_PDBQT,
            " ".join([_OBE.OBABEL_O, self.adv_additional.output_receptor_pdbqt]),
            "".join([_OBE.OBABEL_X, _OBE.OBABEL_X_R]),
            _OBE.OBABEL_P,
            str(self.adv_additional.pH),
            _OBE.OBABEL_PARTIALCHARGE,
            _OBE.OBABEL_PARTIALCHARGE_GASTEIGER,
        ]
        self._openbabel_executor.execute(
            command=_OBE.OBABEL, arguments=arguments, check=True
        )
        self._logger.log(
            f"Exported target as PDBQT file {self.adv_additional.output_receptor_pdbqt}.",
            _LE.INFO,
        )

    def _log_extract_box(self):
        x_coords, y_coords, z_coords = self._extract_box()
        if x_coords is not None:

            def dig(value):
                return round(value, ndigits=2)

            self._logger.log(
                f"Calculating lingad dimensions for AutoDock Vina docking protocol.",
                _LE.INFO,
            )
            self._logger.log(
                f"Ligand ({self.adv_additional.extract_box.reference_ligand_path}):",
                _LE.INFO,
            )
            self._logger_blank.log(
                f"X coordinates: min={dig(min(x_coords))}, max={dig(max(x_coords))}, mean={dig(sum(x_coords) / len(x_coords))}",
                _LE.INFO,
            )
            self._logger_blank.log(
                f"Y coordinates: min={dig(min(y_coords))}, max={dig(max(y_coords))}, mean={dig(sum(y_coords) / len(y_coords))}",
                _LE.INFO,
            )
            self._logger_blank.log(
                f"Z coordinates: min={dig(min(z_coords))}, max={dig(max(z_coords))}, mean={dig(sum(z_coords) / len(z_coords))}",
                _LE.INFO,
            )

    def _extract_box(self):
        # extracts box suggestions from a reference ligand, which can be added to a AutoDock Vina run
        # load the reference file (PDB or SDF)
        ref_format = self.adv_additional.extract_box.reference_ligand_format.upper()
        if ref_format == _STE.EXTRACT_BOX_REFERENCE_LIGAND_FORMAT_PDB:
            ref_mol = Chem.MolFromPDBFile(
                self.adv_additional.extract_box.reference_ligand_path, sanitize=True
            )
        elif ref_format == _STE.EXTRACT_BOX_REFERENCE_LIGAND_FORMAT_SDF:
            mol_supplier = Chem.SDMolSupplier(
                self.adv_additional.extract_box.reference_ligand_path
            )
            for mol in mol_supplier:
                if mol is None:
                    raise StepFailed(
                        f"Could not load molecule from {self.adv_additional.extract_box.reference_ligand_path} - abort."
                    )
                ref_mol = mol
                break
        else:
            raise StepFailed(
                f"Reference ligand format {ref_format} not supported, use PDB or SDF instead - abort."
            )

        # extract coordinates
        x_coords = [atom[0] for atom in ref_mol.GetConformer(0).GetPositions()]
        y_coords = [atom[1] for atom in ref_mol.GetConformer(0).GetPositions()]
        z_coords = [atom[2] for atom in ref_mol.GetConformer(0).GetPositions()]
        return x_coords, y_coords, z_coords

    def execute(self):
        # translate input PDB file into output PDBQT file
        self._export_as_pdb2pdbqt()

        # extract and log the "box" dimensions based on the reference ligand
        self._log_extract_box()



class BaseAgentEnum:

    HEADER = "header"
    STEPS = "steps"

    # header
    # ---------
    ID = "id"
    DESCRIPTION = "description"
    GLOBAL_VARIABLES = "global_variables"
    GLOBAL_SETTINGS = "global_settings"
    VERSION = "version"
    LOGGING = "logging"
    LOGGING_LOGFILE = "logfile"

    # exporting environment variables
    ENVIRONMENT = "environment"
    ENVIRONMENT_EXPORT = "export"
    ENVIRONMENT_EXPORT_KEY = "key"
    ENVIRONMENT_EXPORT_VALUE = "value"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class WorkflowEnum(BaseAgentEnum):

    WORKFLOW = "workflow"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class SchedulerEnum(BaseAgentEnum):

    SCHEDULER = "scheduler"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")



import os
import tempfile
from typing import List

from pydantic import BaseModel
from rdkit import Chem

from icolos.core.workflow_steps.schrodinger.base import StepSchrodingerBase
from icolos.utils.execute_external.ligprep import LigprepExecutor
from icolos.utils.general.files_paths import gen_tmp_file

from icolos.utils.general.molecules import get_charge_for_molecule
from icolos.core.containers.compound import Enumeration, Conformer, get_compound_by_id

from icolos.utils.enums.program_parameters import (
    LigprepEnum,
)
from icolos.utils.enums.step_enums import StepLigprepEnum
from icolos.core.workflow_steps.step import _LE, _CTE
from icolos.utils.general.parallelization import SubtaskContainer, Parallelizer, Subtask
from icolos.utils.general.print_log import print_log_file
from icolos.utils.smiles import to_smiles

_EE = LigprepEnum()
_SLE = StepLigprepEnum()


class StepLigprep(StepSchrodingerBase, BaseModel):
    """
    Interface to the LigPrep binary for ligand embedding
    """

    class Config:
        underscore_attrs_are_private = True

    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executor and test availability
        self._initialize_backend(executor=LigprepExecutor)
        self._check_backend_availability()

    def _prepare_ligprep_arguments(self) -> list:
        arguments_list = []

        # add user-specified command-line settings (if provided); note, that empty dictionaries evaluate
        # to False
        if len(self.settings.arguments.flags) > 0:
            for flag in self.settings.arguments.flags:
                arguments_list.append(str(flag))
        if self.settings.arguments.parameters:
            for key in self.settings.arguments.parameters.keys():
                if key == _EE.LIGPREP_F:
                    self._logger.log(
                        'Removing "-f" parameter for Ligprep arguments - filter file settings need to be specified in the "additional" block directly.',
                        _LE.WARNING,
                    )
                    continue

                arguments_list.append(key)
                if (
                    self.settings.arguments.parameters[key] is not None
                    and self.settings.arguments.parameters[key] != ""
                ):
                    arguments_list.append(str(self.settings.arguments.parameters[key]))

        # add default settings, that are not exposed to the user yet
        if _EE.LIGPREP_HOST not in arguments_list:
            arguments_list.append(_EE.LIGPREP_HOST)
            arguments_list.append(_EE.LIGPREP_HOST_LOCALHOST)
        arguments_list.append(_EE.LIGPREP_WAIT)
        arguments_list = arguments_list + [_EE.LIGPREP_NJOBS, 1]

        return arguments_list

    def _generate_temporary_input_output_files(self, batch: List[List[Subtask]]):
        tmp_output_dirs = []
        tmp_input_smi_paths = []
        tmp_input_filter_paths = []
        tmp_output_sdf_paths = []
        dict_original_smiles = {}

        for next_subtask_list in batch:
            # generate temporary input files and output directory
            cur_tmp_output_dir = tempfile.mkdtemp()
            _, cur_tmp_smi = gen_tmp_file(suffix=".smi", dir=cur_tmp_output_dir)
            _, cur_tmp_filter = gen_tmp_file(suffix=".lff", dir=cur_tmp_output_dir)

            # write smiles to temporary file as "Ligprep" backend
            with open(cur_tmp_smi, "w") as f:
                for subtask in next_subtask_list:
                    enumeration = subtask.data
                    dict_original_smiles[
                        enumeration.get_index_string()
                    ] = enumeration.get_original_smile()
                    f.write(
                        enumeration.get_original_smile()
                        + " "
                        + enumeration.get_index_string()
                        + "\n"
                    )

            # add the path to which "_dock_subjob()" will write the result SDF
            _, output_sdf_path = gen_tmp_file(
                suffix="_result.sdf", dir=cur_tmp_output_dir
            )

            # add the temporary paths
            tmp_output_dirs.append(cur_tmp_output_dir)
            tmp_input_smi_paths.append(cur_tmp_smi)
            tmp_input_filter_paths.append(cur_tmp_filter)
            tmp_output_sdf_paths.append(output_sdf_path)
        return (
            tmp_output_dirs,
            tmp_input_smi_paths,
            tmp_output_sdf_paths,
            tmp_input_filter_paths,
            dict_original_smiles,
        )

    def _execute_ligprep(self):
        ligprep_parallelizer = Parallelizer(func=self._run_subjob)

        # continue until everything is successfully done or number of retries have been exceeded
        while self._subtask_container.done() is False:
            next_batch = self._get_sublists(get_first_n_lists=self._get_number_cores())

            # generate paths and initialize molecules (so that if they fail, this can be covered)
            (
                tmp_output_dirs,
                tmp_input_smi_paths,
                tmp_output_sdf_paths,
                tmp_input_filter_paths,
                dict_original_smiles,
            ) = self._generate_temporary_input_output_files(next_batch)

            # call "token guard" method (only executed, if block is specified in the configuration), which will wait
            # with the execution if not enough tokens are available at the moment
            self._apply_token_guard()

            # execute the current batch in parallel; hand over lists of parameters (will be handled by Parallelizer)
            # also increment the tries and set the status to "failed" (don't do that inside subprocess, as data is
            # copied, not shared!)
            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]
            ligprep_parallelizer.execute_parallel(
                smi_ligand_path=tmp_input_smi_paths,
                path_sdf_results=tmp_output_sdf_paths,
                tmp_output_dir=tmp_output_dirs,
                tmp_input_filter=tmp_input_filter_paths,
                sublist=next_batch,
            )

            # parse the output of that particular batch and remove temporary files
            self._parse_ligprep_output(
                tmp_output_sdf_paths, dict_original_smiles, next_batch
            )
            self._remove_temporary(tmp_output_dirs)

            # print the progress for this execution
            self._log_execution_progress()

    def _parse_ligprep_output(
        self,
        tmp_output_sdf_paths: List[str],
        dict_original_smiles: dict,
        batch: List[List[Subtask]],
    ):
        # TODO: refactor that
        def _update_subtask(sublist: List[Subtask], enum_identifier: str):
            for task in sublist:
                if task.data.get_index_string() == enum_identifier:
                    task.set_status_success()

        for i in range(len(tmp_output_sdf_paths)):
            # get input and output paths and check the files are there
            path_sdf_results = tmp_output_sdf_paths[i]
            cur_sublist = batch[i]
            if (
                not os.path.isfile(path_sdf_results)
                or os.path.getsize(path_sdf_results) == 0
            ):
                continue

            mol_supplier = Chem.SDMolSupplier(path_sdf_results, removeHs=False)
            for mol in mol_supplier:
                # Ligprep adds a "-1" to "-[N]" to the names in the variants tag; this tag is always added
                # alternatively, the "_Name" property could be loaded
                # TODO: add loading only the most likely tautomer here (based on _SLE.LIGPREP_TAUTOMER_PROBABILITY)
                if mol is not None and mol.HasProp(_SLE.LIGPREP_VARIANTS):
                    identifier, _ = mol.GetProp(_SLE.LIGPREP_VARIANTS).split("-")
                    compound_id, enumeration_id = identifier.split(":")
                    compound = get_compound_by_id(
                        self.get_compounds(), int(compound_id)
                    )
                    enumeration = Enumeration(
                        compound_object=compound,
                        smile=to_smiles(mol),
                        original_smile=dict_original_smiles[identifier],
                        molecule=mol,
                    )
                    compound.add_enumeration(enumeration, auto_update=True)
                    _update_subtask(cur_sublist, enum_identifier=identifier)
                else:
                    self._logger.log(
                        f"Skipped molecule when loading as specified property {_SLE.LIGPREP_VARIANTS} could not be found - typically, this indicates that ligprep could not embed the molecule.",
                        _LE.WARNING,
                    )

    def _add_filtering(self, arguments: list, tmp_input_filter: str) -> list:
        filter_file_settings = self.settings.additional.get(_SLE.FILTER_FILE, None)
        if filter_file_settings is not None:
            filter_file = open(tmp_input_filter, "w")
            for key in filter_file_settings.keys():
                filter_file.write(
                    f"{key}                             {filter_file_settings[key]}\n"
                )
            filter_file.close()
            arguments = arguments + [_EE.LIGPREP_F, tmp_input_filter]
        return arguments

    def _run_subjob(
        self,
        smi_ligand_path: str,
        path_sdf_results: str,
        tmp_output_dir: str,
        tmp_input_filter: str,
        sublist: List[Subtask],
    ):
        # 1) increase the sublist "tries" and set status to "failed"
        _ = [task.increment_tries() for task in sublist]
        _ = [task.set_status_failed() for task in sublist]

        # 2) change to directory, to be able to use relative paths (to compensate for Schrodinger bug with AWS)
        working_dir = os.getcwd()
        os.chdir(tmp_output_dir)

        # 3) prepare "Ligprep" arguments
        arguments = self._prepare_ligprep_arguments()
        arguments = self._add_filtering(arguments, tmp_input_filter)
        arguments = arguments + [
            _EE.LIGPREP_INPUT_ISMI,
            os.path.basename(smi_ligand_path),
        ]
        arguments = arguments + [
            _EE.LIGPREP_OUTPUT_OSD,
            os.path.basename(path_sdf_results),
        ]

        # 4) run "Ligprep" backend and add log file to "debug" mode logging
        self._backend_executor.execute(
            command=_EE.LIGPREP,
            arguments=arguments,
            location=tmp_output_dir,
            check=False,
        )

        self._logger.log(
            f"Executed Ligprep backend (output file: {path_sdf_results}).", _LE.DEBUG
        )
        path_tmp_log = os.path.join(
            tmp_output_dir,
            "".join(
                [
                    os.path.splitext(os.path.basename(path_sdf_results))[0],
                    _EE.LIGPREP_LOG_ENDING,
                ]
            ),
        )
        print_log_file(path=path_tmp_log, logger=self._logger, level=_LE.DEBUG)

        # 5) revert back to working directory
        os.chdir(working_dir)

    def _parse_ligprep_result(
        self, sdf_output: str, enumeration: Enumeration
    ) -> List[Conformer]:
        charge = str(
            get_charge_for_molecule(enumeration.get_molecule(), add_as_tag=False)
        )
        mol_supplier = Chem.SDMolSupplier(sdf_output, removeHs=False)
        conformers = []
        for mol_id, mol in enumerate(mol_supplier):
            # note, that formal charge information would be kept if available before (i.e. it retains tags)
            mol.SetProp(_CTE.FORMAL_CHARGE_TAG, charge)
            conformers.append(Conformer(conformer=mol))
        return conformers

    def _log_execution(self, initial_enum_number: int):
        number_enumerations_after = 0
        for compound in self.get_compounds():
            number_enumerations_after += len(compound.get_enumerations())
        self._logger.log(
            f"Executed LigPrep for {initial_enum_number} input enumerations, resulting in {number_enumerations_after} output enumerations.",
            _LE.INFO,
        )
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                self._logger.log(
                    f"Added enumeration {enumeration.get_index_string()} with smile {enumeration.get_smile()}.",
                    _LE.DEBUG,
                )

    def execute(self):
        # in order to be able to efficiently execute Ligprep on the enumeration level, all of them have to be unrolled
        # Note: As they retain their respective Compound object, the attribution later on is simple
        all_enumerations = []
        for compound in self.get_compounds():
            all_enumerations = all_enumerations + compound.get_enumerations()
            compound.clear_enumerations()
        # TODO: we will use the "original_smile" of the enumeration to start the embedding; make sure it exists

        # split into sublists, according to the settings
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(all_enumerations)

        self._execute_ligprep()
        self._log_execution(initial_enum_number=len(all_enumerations))



from abc import ABC, abstractmethod

from icolos.utils.enums.logging_enums import LoggingConfigEnum


class BaseLogger(ABC):
    def __init__(self):
        self._LE = LoggingConfigEnum()
        self._logger = self._initialize_logger()

    def log(self, message: str, level: str):
        if level == self._LE.DEBUG:
            self._logger.debug(message)
        elif level == self._LE.INFO:
            self._logger.info(message)
        elif level == self._LE.WARNING:
            self._logger.warning(message)
        elif level == self._LE.ERROR:
            self._logger.error(message)
        elif level == self._LE.EXCEPTION:
            self._logger.exception(message)
        else:
            raise ValueError("Logger level not supported.")

    @abstractmethod
    def _initialize_logger(self):
        raise NotImplementedError("Overwrite this method in child classes.")



from asyncio import run_coroutine_threadsafe
import subprocess
from subprocess import CompletedProcess
import time
from typing import Callable, Deque, Dict, List
from pydantic import BaseModel
from icolos.core.containers.compound import Compound, Conformer
from icolos.core.containers.perturbation_map import Node, PerturbationMap
from rdkit.Chem import rdmolops
from rdkit import Chem
from icolos.core.containers.compound import Compound, Conformer
from icolos.core.containers.perturbation_map import Node, PerturbationMap
from icolos.core.workflow_steps.step import StepBase
from icolos.utils.enums.parallelization import ParallelizationEnum
from icolos.utils.enums.program_parameters import GromacsEnum, SlurmEnum, StepPMXEnum
from icolos.utils.enums.step_enums import StepGromacsEnum
from icolos.utils.execute_external.execute import Executor
from icolos.utils.execute_external.gromacs import GromacsExecutor
import os
from icolos.utils.general.parallelization import Parallelizer, Subtask
from icolos.core.workflow_steps.step import _LE
import shutil
import glob
from collections import deque

from icolos.utils.general.progress_bar import get_progress_bar_string

_GE = GromacsEnum()
_SGE = StepGromacsEnum()
_SPE = StepPMXEnum()
_PE = ParallelizationEnum
_SE = SlurmEnum


class StepPMXBase(StepBase, BaseModel):

    _antechamber_executor: Executor = None
    _gromacs_executor: Executor = None
    sim_types: List = None
    states: List = None
    therm_cycle_branches: List = None
    run_type: str = None
    ff: str = None
    boxshape: str = None
    boxd: float = None
    water: str = None
    conc: float = None
    pname: str = None
    nname: str = None
    mdp_prefixes: Dict = None

    def __init__(self, **data):
        super().__init__(**data)

        self._antechamber_executor = Executor()
        self._gromacs_executor = GromacsExecutor(
            prefix_execution=self.execution.prefix_execution
        )
        self.sim_types = ["em", "nvt", "eq", "transitions"]
        self.states = ["stateA", "stateB"]
        # for a normal pmx run this would be "water" and "protein"
        # unbound -> ligand, bound -> complex
        self.therm_cycle_branches = ["unbound", "bound"]

        # simulation setup
        self.run_type = self._get_additional_setting(_SPE.RUN_TYPE, "rbfe")
        self.ff = "amber99sb-star-ildn-mut.ff"
        self.boxshape = self._get_additional_setting(_SPE.BOXSHAPE, "dodecahedron")
        self.boxd = self._get_additional_setting(_SPE.BOXD, 1.5)
        self.water = self._get_additional_setting(_SPE.WATER, "tip3p")
        self.conc = self._get_additional_setting(_SPE.CONC, 0.15)
        self.pname = self._get_additional_setting(_SPE.PNAME, "NaJ")
        self.nname = self._get_additional_setting(_SPE.NNAME, "ClJ")
        self.mdp_prefixes = {
            "em": "em",
            "nvt": "nvt",
            "npt": "npt",
            "eq": "eq",
            "transitions": "ti",
        }

    def _get_specific_path(
        self,
        workPath=None,
        edge=None,
        bHybridStrTop=False,
        wp=None,
        state=None,
        r=None,
        sim=None,
    ):
        """
        Utility function for getting the right paths from a pmx-type directory structure.  Works for both rbfe and abfe runs
        """
        if edge == None:
            return workPath
        edgepath = "{0}/{1}".format(workPath, edge)

        if bHybridStrTop == True:
            hybridStrPath = "{0}/hybridStrTop".format(edgepath)
            return hybridStrPath

        if wp == None:
            return edgepath
        wppath = "{0}/{1}".format(edgepath, wp)

        if state == None:
            return wppath
        statepath = "{0}/{1}".format(wppath, state)

        if r == None:
            return statepath
        runpath = "{0}/run{1}".format(statepath, r)

        if sim == None:
            return runpath
        simpath = "{0}/{1}".format(runpath, sim)
        return simpath

    def _parametrise_protein(
        self,
        protein: str = "protein.pdb",
        path: str = "input/protein",
        output="protein.pdb",
    ):
        # run pdb2gmx on the protein
        pdb2gmx_args = [
            "-f",
            os.path.join(self.work_dir, path, protein),
            "-ignh",
            "-water",
            self.settings.additional["water"],
            "-ff",
            self.settings.additional["forcefield"],
            "-o",
            os.path.join(self.work_dir, path, output),
        ]
        self._backend_executor.execute(
            command=_GE.PDB2GMX,
            arguments=pdb2gmx_args,
            check=True,
            location=os.path.join(self.work_dir, path),
        )

    def _prepare_single_tpr(
        self,
        simpath,
        toppath,
        state,
        sim_type,
        executor,
        empath=None,
    ) -> CompletedProcess:
        mdp_path = os.path.join(self.work_dir, "input/mdp")
        mdp_prefix = self.mdp_prefixes[sim_type]

        # TODO: is this a liability? would we ever have more than a single topol file?
        top = "{0}/*.top".format(toppath)
        tpr = "{0}/tpr.tpr".format(simpath)
        mdout = "{0}/mdout.mdp".format(simpath)
        # mdp
        if state == "stateA":
            mdp = "{0}/{1}_l0.mdp".format(mdp_path, mdp_prefix)
        else:
            mdp = "{0}/{1}_l1.mdp".format(mdp_path, mdp_prefix)
        # TODO: deal with nvt/npt for abfe
        # str
        if not sim_type == "transitions":
            if sim_type == "em":
                if self.run_type == "rbfe":
                    inStr = f"{toppath}/ions.pdb"
                elif self.run_type == "abfe":
                    inStr = f"{toppath}/genion.gro"
            elif sim_type in ("eq", "nvt", "npt"):
                inStr = "{0}/confout.gro".format(empath)

            grompp_args = [
                "-f",
                mdp,
                "-c",
                inStr,
                "-r",
                inStr,
                "-p",
                top,
                "-o",
                tpr,
                "-maxwarn",
                4,
                "-po",
                mdout,
            ]
            if not os.path.isfile(tpr):
                result = executor.execute(
                    command=_GE.GROMPP,
                    arguments=grompp_args,
                    check=True,
                    location=simpath,
                )
            else:
                self._logger.log(f"tpr file {tpr} already exists, skipping", _LE.DEBUG)

        elif sim_type == "transitions":
            grompp_full_cmd = []
            # 80 frames = 0 - 79
            num_frames = len([f for f in os.listdir(simpath) if f.startswith("frame")])
            self._logger.log(
                f"Generating transition tpr files for {num_frames} frames", _LE.DEBUG
            )
            for frame in range(num_frames):
                inStr = f"{simpath}/frame{frame}.gro"
                tpr = f"{simpath}/ti{frame}.tpr".format(simpath, frame)

                grompp_args = [
                    "gmx grompp",
                    "-f",
                    mdp,
                    "-c",
                    inStr,
                    "-r",
                    inStr,
                    "-p",
                    top,
                    "-o",
                    tpr,
                    "-maxwarn",
                    "4",
                    "-po",
                    mdout,
                    ";",
                ]
                if not os.path.isfile(tpr):
                    grompp_full_cmd += grompp_args
                else:
                    self._logger.log(
                        f"tpr file {tpr} already exists, skipping", _LE.DEBUG
                    )
            grompp_full_cmd = " ".join(grompp_full_cmd[:-1])
            # check all transitions have not been skipped
            if grompp_full_cmd:
                result = executor.execute(
                    command=grompp_full_cmd, arguments=[], check=True, location=simpath
                )
        self._clean_backup_files(simpath)

    def _clean_pdb_structure(self, tmp_dir: str) -> None:
        files = [file for file in os.listdir(tmp_dir) if file.endswith("pdb")]
        for file in files:
            cleaned_lines = []
            with open(os.path.join(tmp_dir, file), "r") as f:
                lines = f.readlines()
            for line in lines:
                if "ATOM" in line or "HETATM" in line:
                    cleaned_lines.append(line)
            with open(os.path.join(tmp_dir, file), "w") as f:
                f.writelines(cleaned_lines)

    def _parametrisation_pipeline(
        self, tmp_dir, conf: Conformer, include_top=False, include_gro=False
    ):
        # main pipeline for producing GAFF parameters for a ligand
        charge_method = self._get_additional_setting(
            key=_SGE.CHARGE_METHOD, default="bcc"
        )
        formal_charge = (
            rdmolops.GetFormalCharge(conf.get_molecule()) if conf is not None else 0
        )
        arguments_acpype = [
            "-di",
            "MOL.sdf",
            "-c",
            charge_method,
            "-a",
            "gaff2",
            "-n",
            formal_charge,
        ]
        self._logger.log("Generating ligand parameters...", _LE.DEBUG)
        self._backend_executor.execute(
            command=_GE.ACPYPE_BINARY,
            arguments=arguments_acpype,
            location=tmp_dir,
            check=True,
        )
        # search the output dir for the itp file
        acpype_dir = [p for p in os.listdir(tmp_dir) if p.endswith(".acpype")][0]
        itp_file = [
            f
            for f in os.listdir(os.path.join(tmp_dir, acpype_dir))
            if f.endswith("GMX.itp")
        ][0]
        pdb_file = [
            f
            for f in os.listdir(os.path.join(tmp_dir, acpype_dir))
            if f.endswith("NEW.pdb")
        ][0]
        shutil.copyfile(
            os.path.join(tmp_dir, acpype_dir, itp_file),
            # standardized name must be enforced here to make argument
            # parsing easier in subsequent pmx steps
            os.path.join(tmp_dir, "MOL.itp"),
        )
        shutil.copyfile(
            os.path.join(tmp_dir, acpype_dir, pdb_file),
            # standardized name must be enforced here to make argument
            # parsing easier in subsequent pmx steps
            os.path.join(tmp_dir, "MOL.pdb"),
        )
        # for abfe calculations we need the ligand_GMX.top + .gro files as well
        if include_top:
            top_file = [
                f
                for f in os.listdir(os.path.join(tmp_dir, acpype_dir))
                if f.endswith("GMX.top")
            ][0]
            shutil.copyfile(
                os.path.join(tmp_dir, acpype_dir, top_file),
                os.path.join(tmp_dir, top_file),
            )
        if include_gro:
            gro_file = [
                f
                for f in os.listdir(os.path.join(tmp_dir, acpype_dir))
                if f.endswith("GMX.gro")
            ][0]
            shutil.copyfile(
                os.path.join(tmp_dir, acpype_dir, gro_file),
                os.path.join(tmp_dir, gro_file),
            )

    def _run_job_pool(self, run_func: Callable):
        # get the loaded tasks from the subtask container

        # while self._subtask_container.done() is False:
        job_generator = (j for j in self._subtask_container.get_todo_tasks())
        n_jobs = len(self._subtask_container.get_todo_tasks())
        current_jobs = []
        # initially fill the queue with N jobs
        while len(current_jobs) < self.execution.parallelization.jobs:
            try:
                current_jobs.append(next(job_generator))
            except StopIteration:
                break

        _ = [job.increment_tries() for job in current_jobs]
        # submit the initial job pool
        queue_exhausted = False
        previous_metrics = [0, 0, 0]
        done_count = 0
        while done_count < n_jobs:
            # loop through the jobs:
            done_count = len(self._subtask_container.get_done_tasks())
            running_count = len(self._subtask_container.get_running_tasks())
            ready_count = len(self._subtask_container.get_todo_tasks())

            current_metrics = [done_count, running_count, ready_count]
            if current_metrics != previous_metrics:
                self._logger.log(
                    f" Execution Summary: PENDING: {ready_count}\tRUNNING: {running_count}\tDONE: {done_count}",
                    _LE.INFO,
                )
                prog_string = get_progress_bar_string(
                    done_count, done_count + running_count + ready_count
                )
                self._logger.log(prog_string, _LE.INFO)
            previous_metrics = current_metrics
            for job in current_jobs:
                # job is ready to go, dispatch it to Slurm
                if job.status == _PE.STATUS_READY:
                    job_id = run_func(job.data)
                    job.set_job_id(job_id)
                    job.set_status(_PE.STATUS_RUNNING)
                # check the job status
                elif job.status == _PE.STATUS_RUNNING:
                    # check to see whether it's finished
                    status = self._backend_executor._check_job_status(job.job_id)
                    if status == _SE.COMPLETED:
                        self._logger.log(f"Job {job.job_id} COMPLETED", _LE.DEBUG)
                        job.set_status_success()
                    elif status == _SE.FAILED:
                        self._logger.log(f"Job {job.job_id} FAILED!", _LE.WARNING)
                        job.set_status_failed()
                    elif status == _SE.CANCELLED:
                        self._logger.log(
                            f"Job {job.job_id} was CANCELLED!", _LE.WARNING
                        )
                        job.set_status_failed()
                    elif status == _SE.NODE_FAIL:
                        # aws revoked the spot instance.  Resubmit the job
                        self._logger.log(
                            f"Job {job.job_id} was revoked, resubmitting...", _LE.DEBUG
                        )
                        job.set_status(_PE.STATUS_READY)
                    elif status not in (_SE.RUNNING, _SE.PENDING):
                        self._logger.log(
                            f"Unhandled job state {status} for job {job.job_id}",
                            _LE.WARNING,
                        )
                        job.set_status_failed()

                # if complete, succesfully or not, remove the job from the queue, prepare another
                elif job.status in (_PE.STATUS_SUCCESS, _PE.STATUS_FAILED):
                    current_jobs.remove(job)
                    if queue_exhausted is False:
                        try:
                            new_job = next(job_generator)
                            self._logger.log(f"Preparing new job {job.data}", _LE.DEBUG)
                            new_job.increment_tries()
                            current_jobs.append(new_job)
                        except StopIteration:
                            self._logger.log("Reached end of job queue", _LE.DEBUG)
                            queue_exhausted = True
            time.sleep(10)

    def _execute_pmx_step_parallel(
        self,
        run_func: Callable,
        step_id: str,
        result_checker: Callable,
        prune_completed: bool = True,
        **kwargs,
    ):
        """
        Instantiates Icolos's parallelizer object,
        runs the step's execute method,
        passes any kwargs straight to the run_func
        If result_checker is provided,
        """
        parallelizer = Parallelizer(func=run_func)
        n = 1
        while self._subtask_container.done() is False:

            next_batch = self._get_sublists(
                get_first_n_lists=self._get_number_cores()
            )  # return n lists of length max_sublist_length
            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]

            jobs = self._prepare_edges(next_batch)
            n_removed = 0
            if prune_completed:
                pre_exec_results = result_checker(jobs)
                for job_sublist, exec_success_sublist, sublist in zip(
                    jobs, pre_exec_results, next_batch
                ):
                    # we test on the subtask level, not the individual job level, but since jobs are run through with max_len_sublists=1, in practice this doesn't matter
                    for job, result, task in zip(
                        job_sublist, exec_success_sublist, sublist
                    ):
                        if result is True:
                            # remove the entire sublist (one fewer cores running)
                            job_sublist.remove(job)
                            task.set_status_success()
                            self._logger.log(
                                f"Removed job {job} from execution batch, good output found",
                                _LE.DEBUG,
                            )
                            n_removed += 1
                        # if we have emptied entire job queues, remove the queue
                self._logger.log(
                    f"Executing {step_id} for batch {n}, containing {len(jobs)} * {self.execution.parallelization.max_length_sublists} jobs",
                    _LE.INFO,
                )

            jobs = [j for j in jobs if j]
            parallelizer.execute_parallel(jobs=jobs, **kwargs)

            self._logger.log("Checking execution results...", _LE.DEBUG)
            batch_results = result_checker(jobs)
            good_results = 0
            for task, result in zip(next_batch, batch_results):
                # returns boolean arrays: False => failed job
                for subtask, sub_result in zip(task, result):
                    if sub_result == False:
                        subtask.set_status_failed()
                        self._logger.log(f"Warning: job {subtask} failed!", _LE.WARNING)
                        if (
                            self.get_perturbation_map() is not None
                            and self.get_perturbation_map().strict_execution
                            and isinstance(subtask.data, str)
                        ):
                            edge = self.get_perturbation_map().get_edge_by_id(
                                subtask.data
                            )
                            if edge is not None:
                                edge._set_status(_PE.STATUS_FAILED)

                    else:
                        subtask.set_status_success()
                        good_results += 1

            self._logger.log(
                f"EXECUTION SUMMARY: Completed {good_results} jobs successfully (out of {len(next_batch) * len(next_batch[0])} jobs for step {step_id}. Removed {n_removed} already completed jobs",
                _LE.INFO,
            )

            self._log_execution_progress()
            n += 1

    def get_edges(self):
        """
        Inspect the map object  passed to the step and extract the edge info
        """

        return self.get_workflow_object().workflow_data.perturbation_map.edges

    def get_nodes(self):
        """
        return the nodes attached to the perturbation map
        """
        return self.get_workflow_object().workflow_data.perturbation_map.nodes

    def _get_line_idx(self, data: list, id_str: str) -> int:
        line = [e for e in data if id_str in e]
        assert len(line) == 1
        line = line[0]
        return data.index(line)

    def _clean_protein(self):
        existing_itp_files = [
            f
            for f in os.listdir(os.path.join(self.work_dir, "input/protein"))
            if f.endswith("itp") and "Protein" in f
        ]
        if (
            not existing_itp_files
        ):  # no protein itp files, we have a single chain that needs extacting from the top file
            with open(os.path.join(self.work_dir, "input/protein/topol.top"), "r") as f:
                top_lines = f.readlines()

            moltype_line = self._get_line_idx(top_lines, _GE.MOLECULETYPES)

            end_itp_line = self._get_line_idx(top_lines, "; Include water topology")

            moltype = top_lines[moltype_line + 2].split()[0]
            cleaned_top = (
                top_lines[:moltype_line]
                + [f'#include "topol_{moltype}.itp']
                + top_lines[end_itp_line:]
            )

            itp_lines = top_lines[moltype_line:end_itp_line]

            with open(os.path.join(self.work_dir, "input/protein/topol.top"), "w") as f:
                f.writelines(cleaned_top)

            with open(
                os.path.join(self.work_dir, f"input/protein/topol_{moltype}.itp"), "w"
            ) as f:
                f.writelines(itp_lines)

    def get_hub_conformer(self, hub_conf_path) -> Conformer:
        """

        :return _type_: _description_
        """

        with Chem.SDMolSupplier(hub_conf_path) as supplier:
            hub_mol = supplier[0]
        return Conformer(conformer=hub_mol)

    def _construct_perturbation_map(self, work_dir: str, replicas: int):
        if self.get_perturbation_map() is not None:
            self._logger.log("Perturbation map already constructed", _LE.DEBUG)
            self.get_perturbation_map().protein = (
                self.data.generic.get_argument_by_extension("pdb", rtn_file_object=True)
            )
            self.get_perturbation_map().replicas = replicas
            return
        topology = self._get_additional_setting("topology", default="normal")
        # check whether a hub conformer has been supplied (as an sdf file)
        hub_conf_path = self._get_additional_setting("hub_conformer", default=None)

        if hub_conf_path is not None:
            assert hub_conf_path.endswith(
                ".sdf"
            ), "Hub conformer must be supplied as an SDF file!"

        perturbation_map = PerturbationMap(
            compounds=self.data.compounds,
            protein=self.data.generic.get_argument_by_extension(
                "pdb", rtn_file_object=True
            ),
            replicas=replicas,
            strict_execution=self._get_additional_setting(_SPE.STRICT, default=True),
            hub_conformer=self.get_hub_conformer(hub_conf_path)
            if hub_conf_path is not None
            else None,
        )
        if topology == "normal":
            # construct the perturbation map and load in the log file
            log_file = self.data.generic.get_argument_by_extension(
                "log", rtn_file_object=True
            )
            log_file.write(work_dir)

            perturbation_map.parse_map_file(
                os.path.join(self.work_dir, log_file.get_file_name())
            )
        elif topology == "star":
            # manually generate star top, no mapping tool required
            perturbation_map.generate_star_map()

        self._logger.log(
            f"Initialised perturbation map with {len(perturbation_map.get_nodes())} nodes and {len(perturbation_map.get_edges())} edges",
            _LE.INFO,
        )
        self.get_workflow_object().set_perturbation_map(perturbation_map)

    def _prepare_edges(self, batch) -> List[List[str]]:
        edges = []

        for task in batch:
            task_edges = []
            for element in task:
                task_edges.append(element.data)
            edges.append(task_edges)
        return edges

    def _log_result(self, result: CompletedProcess):
        for line in result.stderr.split("\n"):
            self._logger_blank.log(line, _LE.DEBUG)

    def _clean_backup_files(self, path):
        toclean = glob.glob("{0}/*#".format(path))
        for clean in toclean:
            os.remove(clean)

    def _separate_atomtypes(self, lig_path: str) -> None:
        with open(os.path.join(lig_path, "MOL.itp"), "r") as f:
            itp_lines = f.readlines()

        start_idx = self._get_line_idx(itp_lines, _GE.ATOMTYPES)
        stop_index = self._get_line_idx(itp_lines, _GE.MOLECULETYPES)

        atomtype_lines = itp_lines[start_idx:stop_index]
        cleaned_itp_lines = itp_lines[stop_index:]
        with open(os.path.join(lig_path, "MOL.itp"), "w") as f:
            f.writelines(cleaned_itp_lines)

        # process the atomtype lines to remove the bondtype
        # col causes gmx to complain
        cleaned_atomtype_lines = []
        for line in atomtype_lines:
            parts = line.split()
            if len(parts) > 5:
                cleaned_parts = [parts[0]] + parts[2:] + ["\n"]
                cleaned_atomtype_lines.append(" ".join(cleaned_parts))
        with open(os.path.join(lig_path, "ffMOL.itp"), "w") as f:
            f.writelines(cleaned_atomtype_lines)

    def _parametrise_nodes(self, jobs):
        if isinstance(jobs, list):
            node = jobs[0]
        else:
            node = jobs
        if isinstance(node, Node):
            node_id = node.get_node_hash()
            conf = node.conformer
        elif isinstance(node, Compound):
            # in abfe we pass compounds here not edges
            node_id = node.get_index_string()
            conf = node.get_enumerations()[0].get_conformers()[0]
        else:
            raise NotImplementedError(f"Cannot parametrize object of type {type(node)}")
        lig_path = os.path.join(self.work_dir, "input", "ligands", node_id)
        os.makedirs(lig_path, exist_ok=True)
        conf.write(os.path.join(lig_path, "MOL.sdf"))

        # now run ACPYPE on the ligand to produce the topology file
        self._parametrisation_pipeline(lig_path, conf=conf)

        # produces MOL.itp, need to separate the atomtypes directive out into ffMOL.itp for pmx
        # to generate the forcefield later
        self._separate_atomtypes(lig_path)



from icolos.utils.enums.program_parameters import GromacsEnum
from icolos.core.workflow_steps.gromacs.base import StepGromacsBase
from icolos.utils.execute_external.gromacs import GromacsExecutor
from pydantic import BaseModel
from icolos.core.workflow_steps.step import _LE
import os

_GE = GromacsEnum()


class StepGMXrmsd(StepGromacsBase, BaseModel):
    """
    Run gromacs rmsd calculation on trajectory
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=GromacsExecutor)
        self._check_backend_availability()

    def execute(self):

        tmp_dir = self._make_tmpdir()

        # we're going to get a trajectory from a trjconv step, and a single structure to
        # compare against. rmsd is computed for every step of the trj file

        # write out generic files
        self.write_input_files(tmp_dir)

        # conformer coming from a Compound object
        conf = self._unroll_compounds(self.data.compounds)

        conf = conf[0]
        conf.write(os.path.join(tmp_dir, "reference.sdf"), format_="pdb")

        flag_dict = {
            "-s": "reference.pdb",
            "-f": self.data.generic.get_argument_by_extension("xtc"),
            "-fit": "rot+trans",
        }

        arguments = self._parse_arguments(flag_dict=flag_dict, args=["-w"])
        result = self._backend_executor.execute(
            command=_GE.RMS,
            arguments=arguments,
            location=tmp_dir,
            check=True,
            pipe_input='echo -e "2\n2\n"',
        )
        for line in result.stdout.split("\n"):
            self._logger_blank.log(line, _LE.DEBUG)

        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



from typing import Dict, List
from icolos.core.containers.perturbation_map import Edge
from icolos.core.workflow_steps.pmx.base import StepPMXBase
from pydantic import BaseModel
from icolos.utils.execute_external.pmx import PMXExecutor
import os
from icolos.utils.enums.program_parameters import PMXEnum, PMXAtomMappingEnum
from icolos.utils.general.parallelization import SubtaskContainer

_PE = PMXEnum()
_PAE = PMXAtomMappingEnum()


class StepPMXatomMapping(StepPMXBase, BaseModel):
    """Ligand alchemy: map atoms for morphing."""

    def __init__(self, **data):
        super().__init__(**data)
        self._initialize_backend(executor=PMXExecutor)

    def _prepare_arguments(self, args, output_dir):
        # prepare the final set of arguments as a list
        prepared_args = []
        default_args = {
            "-o1": f"{output_dir}/pairs1.dat",
            "-o2": f"{output_dir}/pairs2.dat",
            "-opdb1": f"{output_dir}/out_pdb1.pdb",
            "-opdb2": f"{output_dir}/out_pdb2.pdb",
            "-opdbm1": f"{output_dir}/out_pdbm1.pdb",
            "-opdbm2": f"{output_dir}/out_pdbm2.pdb",
            "-score": f"{output_dir}/score.dat",
            "-log": f"{output_dir}/mapping.log",
        }
        for key, value in args.items():
            default_args[key] = value

        for key, value in default_args.items():
            prepared_args.append(key),
            prepared_args.append(value)
        return prepared_args

    def _execute_command(self, jobs: List[str]):

        edge = jobs[0]
        parts = edge.split("_")
        lig1 = parts[0]
        lig2 = parts[1]
        # write them to the right dir as a pdb from the outset
        arguments = {
            "-i1": os.path.join(
                self.work_dir,
                _PAE.LIGAND_DIR,
                lig1,
                "MOL.pdb",
            ),
            "-i2": os.path.join(
                self.work_dir,
                _PAE.LIGAND_DIR,
                lig2,
                "MOL.pdb",
            ),
        }
        output_dir = os.path.join(self.work_dir, edge, _PE.HYBRID_STR_TOP)
        arguments = self._prepare_arguments(args=arguments, output_dir=output_dir)

        self._backend_executor.execute(
            command=_PE.ATOMMAPPING,
            arguments=arguments,
            check=True,
            location=self.work_dir,
        )

    def execute(self):
        # check the workflow has been configured correctly to use a shared work_dir
        assert self.work_dir is not None and os.path.isdir(self.work_dir)

        edges = [e.get_edge_id() for e in self.get_edges()]
        # enforce single edge per job queue
        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(edges)
        self._execute_pmx_step_parallel(
            run_func=self._execute_command,
            step_id="atomMapping",
            result_checker=self._check_result,
        )

    def _check_result(self, batch: List[List[Edge]]) -> List[List[bool]]:
        """
        Look in each hybridStrTop dir and check the output pdb files exist for the edges
        """
        output_files = ["pairs1.dat", "pairs2.dat", "out_pdb1.pdb", "out_pdb2.pdb"]
        results = []
        for subjob in batch:
            subjob_results = []
            for job in subjob:
                subjob_results.append(
                    all(
                        [
                            os.path.isfile(
                                os.path.join(self.work_dir, job, "hybridStrTop", f)
                            )
                            for f in output_files
                        ]
                    )
                )
            results.append(subjob_results)
        return results



from icolos.utils.enums.program_parameters import PMXEnum
from icolos.utils.execute_external.execute import ExecutorBase

_PE = PMXEnum()


class PMXExecutor(ExecutorBase):
    def __init__(self, prefix_execution=None, binary_location=None):
        super().__init__(
            prefix_execution=prefix_execution, binary_location=binary_location
        )

    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):
        if command not in [
            _PE.ABFE,
            _PE.ANALYSE,
            _PE.ATOMMAPPING,
            _PE.DOUBLEBOX,
            _PE.GENLIB,
            _PE.GENTOP,
            _PE.LIGANDHYBRID,
            _PE.MUTATE,
            _PE.BOX_WATER_IONS,
            _PE.PREPARE_SIMULATIONS,
            _PE.PREPARE_TRANSITIONS,
            _PE.RUN_ANALYSIS,
            _PE.RUN_SIMULATIONS,
            _PE.ASSEMBLE_SYSTEMS,
        ]:
            raise ValueError(
                "Command must be present in internal list of PMX executables."
            )

        # handle for dealing with programs that want interactive input
        return super().execute(
            command=command,
            arguments=arguments,
            check=check,
            location=location,
            pipe_input=pipe_input,
        )

    def is_available(self):
        try:
            result = self.execute(
                command=_PE.ANALYSE, arguments=[_PE.ANALYSE_HELP], check=False
            )
            if _PE.ANALYSE_HELP_SUCCESS_STRING in result.stdout:
                return True
            return False
        except Exception as e:
            return False



from icolos.utils.enums.program_parameters import XTBEnum
from icolos.utils.execute_external.execute import ExecutorBase

EE = XTBEnum()


class XTBExecutor(ExecutorBase):
    """For the execution of the "xtb" binary."""

    def __init__(self, prefix_execution=None, binary_location=None):
        super().__init__(
            prefix_execution=prefix_execution, binary_location=binary_location
        )

    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):
        # check, whether a proper executable is provided
        if command not in [EE.XTB]:
            raise ValueError(
                "Parameter command must be an dictionary of the internal XTB executable list."
            )

        return super().execute(
            command=command,
            arguments=arguments,
            check=check,
            location=None,
            pipe_input=pipe_input,
        )

    def is_available(self):
        try:
            result = self.execute(command=EE.XTB, arguments=[EE.XTB_HELP], check=True)

            if EE.XTB_HELP_IDENTIFICATION_STRING in result.stderr:
                return True
            return False
        except Exception as e:
            return False



from enum import Enum


class ExecutionPlatformEnum(str, Enum):
    LOCAL = "local"
    SLURM = "slurm"
    PARTITION = "partition"
    TIME = "time"
    GRES = "gres"
    MEM = "mem"
    CORES = "cores"
    CORE = "core"
    GPU = "gpu"
    SERIAL = "serial"
